"""
This module provides a function to compute a cube parametrization

"""

import logging
import os

import nibabel as nb
import numpy as np
from lapy import Solver, TetMesh, TriaMesh, io
from scipy import sparse as sp
from scipy import stats as st
from sklearn.decomposition import PCA

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS

# ------------------------------------------------------------------------------
# AUXILIARY FUNCTIONS
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# getSeam()


def getSeam(v4c, t4c, i4c, k4c, v4cBndOpenKeep, t4cBndOpen, anisoLaplEvec):
    # --------------------------------------------------------------------------
    # getSeam subfunctions
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    # getSeam subfunction #1

    # case 1: one nan in the tetra
    def getSeamCase1(v4c, t4c, i4c, k4c, vfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i):
        # do we have one or two neg values?
        # case 1a: one neg value
        if np.sum(vfuncXEv1[t4c[i, :]] < 0) == 1:
            # get indices
            idxPos1 = np.where(vfuncXEv1[t4c[i, :]] > 0)[0][0]
            idxPos2 = np.where(vfuncXEv1[t4c[i, :]] > 0)[0][1]
            idxNeg1 = np.where(vfuncXEv1[t4c[i, :]] < 0)[0][0]
            idxNan1 = np.where(np.isnan(vfuncXEv1[t4c[i, :]]))[0][0]

            # --------------------------------------------------------------
            # get new point 1 (we go neg-->pos)

            # scaling factor
            sf1 = np.abs(vfuncXEv1[t4c[i, idxNeg1]]) / (vfuncXEv1[t4c[i, idxPos1]] - vfuncXEv1[t4c[i, idxNeg1]])

            # difference vector
            dv1 = v4c[t4c[i, idxPos1], :] - v4c[t4c[i, idxNeg1], :]

            # compute new point
            pt1 = sf1 * dv1 + v4c[t4c[i, idxNeg1], :]

            # --------------------------------------------------------------
            # get new point 2

            # scaling factor
            sf2 = np.abs(vfuncXEv1[t4c[i, idxNeg1]]) / (vfuncXEv1[t4c[i, idxPos2]] - vfuncXEv1[t4c[i, idxNeg1]])

            # difference vector
            dv2 = v4c[t4c[i, idxPos2], :] - v4c[t4c[i, idxNeg1], :]

            # compute new point
            pt2 = sf2 * dv2 + v4c[t4c[i, idxNeg1], :]

            # --------------------------------------------------------------
            # store and connect

            # check for duplicate points (pt1)
            if newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]] == 0:
                # add to list
                v4c = np.concatenate((v4c, np.array(pt1, ndmin=2)))
                # keep index
                idxPt1 = len(v4c) - 1
                # update adj mtx
                newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]] = idxPt1
                newVtcsAdj[t4c[i, idxNeg1], t4c[i, idxPos1]] = idxPt1
                # also append i4c and k4c
                i4c = np.concatenate((i4c, st.mode(i4c[t4c[i, :]])[0]))
                k4c = np.concatenate((k4c, st.mode(k4c[t4c[i, :]])[0]))
                # let's remember the new indices, and whether or not it's a
                # 'positive' or 'negative' tetra (in terms of the 2nd
                # eigenfunction)
                newVtcs = np.concatenate((newVtcs, [idxPt1]))
                tmpSgn = np.unique(vfuncXEv2[t4c[i, :]][np.logical_not(np.isnan(vfuncXEv2[t4c[i, :]]))])
                if len(tmpSgn) == 1:
                    newVtcsSgn = np.concatenate((newVtcsSgn, tmpSgn))
                else:
                    newVtcsSgn = np.concatenate((newVtcsSgn, np.array([0])))
                    raise RuntimeError("Inconsistency while creating seam 1, exiting.")
            else:
                idxPt1 = newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]]

            # check for duplicate points (pt2)
            if newVtcsAdj[t4c[i, idxPos2], t4c[i, idxNeg1]] == 0:
                # add to list
                v4c = np.concatenate((v4c, np.array(pt2, ndmin=2)))
                # keep index
                idxPt2 = len(v4c) - 1
                # update adj mtx
                newVtcsAdj[t4c[i, idxPos2], t4c[i, idxNeg1]] = idxPt2
                newVtcsAdj[t4c[i, idxNeg1], t4c[i, idxPos2]] = idxPt2
                # also append i4c and k4c
                i4c = np.concatenate((i4c, st.mode(i4c[t4c[i, :]])[0]))
                k4c = np.concatenate((k4c, st.mode(k4c[t4c[i, :]])[0]))
                # let's remember the new indices, and whether or not it's a
                # 'positive' or 'negative' tetra (in terms of the 2nd
                # eigenfunction)
                newVtcs = np.concatenate((newVtcs, [idxPt2]))
                tmpSgn = np.unique(vfuncXEv2[t4c[i, :]][np.logical_not(np.isnan(vfuncXEv2[t4c[i, :]]))])
                if len(tmpSgn) == 1:
                    newVtcsSgn = np.concatenate((newVtcsSgn, tmpSgn))
                else:
                    newVtcsSgn = np.concatenate((newVtcsSgn, np.array([0])))
                    raise RuntimeError("Inconsistency while creating seam 2, exiting.")
            else:
                idxPt2 = newVtcsAdj[t4c[i, idxPos2], t4c[i, idxNeg1]]

            # create new tetras
            # Neg1, pt1, pt2, nan
            # pt1, pt2, Pos1, nan
            # t2, Pos1, Pos2, nan

            newTetra = np.concatenate((newTetra, np.array((t4c[i, idxNeg1], idxPt1, idxPt2, t4c[i, idxNan1]), ndmin=2)))
            newTetra = np.concatenate((newTetra, np.array((idxPt1, idxPt2, t4c[i, idxPos1], t4c[i, idxNan1]), ndmin=2)))
            newTetra = np.concatenate(
                (newTetra, np.array((idxPt2, t4c[i, idxPos1], t4c[i, idxPos2], t4c[i, idxNan1]), ndmin=2))
            )

        # case 1b: two neg value
        elif np.sum(vfuncXEv1[t4c[i, :]] < 0) == 2:
            # get indices
            idxPos1 = np.where(vfuncXEv1[t4c[i, :]] > 0)[0][0]
            idxNeg1 = np.where(vfuncXEv1[t4c[i, :]] < 0)[0][0]
            idxNeg2 = np.where(vfuncXEv1[t4c[i, :]] < 0)[0][1]
            idxNan1 = np.where(np.isnan(vfuncXEv1[t4c[i, :]]))[0][0]

            # --------------------------------------------------------------
            # get new point 1

            # scaling factor
            sf1 = np.abs(vfuncXEv1[t4c[i, idxNeg1]]) / (vfuncXEv1[t4c[i, idxPos1]] - vfuncXEv1[t4c[i, idxNeg1]])

            # difference vector
            dv1 = v4c[t4c[i, idxPos1], :] - v4c[t4c[i, idxNeg1], :]

            # compute new point
            pt1 = sf1 * dv1 + v4c[t4c[i, idxNeg1], :]

            # --------------------------------------------------------------
            # get new point 2

            # scaling factor
            sf2 = np.abs(vfuncXEv1[t4c[i, idxNeg2]]) / (vfuncXEv1[t4c[i, idxPos1]] - vfuncXEv1[t4c[i, idxNeg2]])

            # difference vector
            dv2 = v4c[t4c[i, idxPos1], :] - v4c[t4c[i, idxNeg2], :]

            # compute new point
            pt2 = sf2 * dv2 + v4c[t4c[i, idxNeg2], :]

            # --------------------------------------------------------------
            # store and connect

            # check for duplicate points (pt1)
            if newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]] == 0:
                # add to list
                v4c = np.concatenate((v4c, np.array(pt1, ndmin=2)))
                # keep index
                idxPt1 = len(v4c) - 1
                # update adj mtx
                newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]] = idxPt1
                newVtcsAdj[t4c[i, idxNeg1], t4c[i, idxPos1]] = idxPt1
                # also append i4c and k4c
                i4c = np.concatenate((i4c, st.mode(i4c[t4c[i, :]])[0]))
                k4c = np.concatenate((k4c, st.mode(k4c[t4c[i, :]])[0]))
                # let's remember the new indices, and whether or not it's a
                # 'positive' or 'negative' tetra (in terms of the 2nd
                # eigenfunction)
                newVtcs = np.concatenate((newVtcs, [idxPt1]))
                tmpSgn = np.unique(vfuncXEv2[t4c[i, :]][np.logical_not(np.isnan(vfuncXEv2[t4c[i, :]]))])
                if len(tmpSgn) == 1:
                    newVtcsSgn = np.concatenate((newVtcsSgn, tmpSgn))
                else:
                    newVtcsSgn = np.concatenate((newVtcsSgn, np.array([0])))
                    raise RuntimeError("Inconsistency while creating seam 3, exiting.")
            else:
                idxPt1 = newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]]

            # check for duplicate points (pt2)
            if newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg2]] == 0:
                # add to list
                v4c = np.concatenate((v4c, np.array(pt2, ndmin=2)))
                # keep index
                idxPt2 = len(v4c) - 1
                # update adj mtx
                newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg2]] = idxPt2
                newVtcsAdj[t4c[i, idxNeg2], t4c[i, idxPos1]] = idxPt2
                # also append i4c and k4c
                i4c = np.concatenate((i4c, st.mode(i4c[t4c[i, :]])[0]))
                k4c = np.concatenate((k4c, st.mode(k4c[t4c[i, :]])[0]))
                # let's remember the new indices, and whether or not it's a
                # 'positive' or 'negative' tetra (in terms of the 2nd
                # eigenfunction)
                newVtcs = np.concatenate((newVtcs, [idxPt2]))
                tmpSgn = np.unique(vfuncXEv2[t4c[i, :]][np.logical_not(np.isnan(vfuncXEv2[t4c[i, :]]))])
                if len(tmpSgn) == 1:
                    newVtcsSgn = np.concatenate((newVtcsSgn, tmpSgn))
                else:
                    newVtcsSgn = np.concatenate((newVtcsSgn, np.array([0])))
                    raise RuntimeError("Inconsistency while creating seam 4, exiting.")
            else:
                idxPt2 = newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg2]]

            # create new tetras
            # pos1, pt1, pt2, nan
            # pt1, pt2, neg1, nan
            # pt2, neg1, neg2, nan

            newTetra = np.concatenate((newTetra, np.array((t4c[i, idxPos1], idxPt1, idxPt2, t4c[i, idxNan1]), ndmin=2)))
            newTetra = np.concatenate((newTetra, np.array((idxPt1, idxPt2, t4c[i, idxNeg1], t4c[i, idxNan1]), ndmin=2)))
            newTetra = np.concatenate(
                (newTetra, np.array((idxPt2, t4c[i, idxNeg1], t4c[i, idxNeg2], t4c[i, idxNan1]), ndmin=2))
            )

        else:
            raise RuntimeError("Inconsistency while creating seam (incorrect number of negative values), exiting.")

        # return
        return v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn

    # --------------------------------------------------------------------------
    # getSeam subfunction #2

    # case 2: two nans in tetra
    def getSeamCase2(v4c, t4c, i4c, k4c, vfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i):
        # get indices
        idxPos1 = np.where(vfuncXEv1[t4c[i, :]] > 0)[0][0]
        idxNeg1 = np.where(vfuncXEv1[t4c[i, :]] < 0)[0][0]
        idxNan1 = np.where(np.isnan(vfuncXEv1[t4c[i, :]]))[0][0]
        idxNan2 = np.where(np.isnan(vfuncXEv1[t4c[i, :]]))[0][1]

        # --------------------------------------------------------------
        # get new point
        np.where(vfuncXEv1[t4c[i, :]] < 0)[0][0]
        np.where(vfuncXEv1[t4c[i, :]] > 0)[0][0]

        # scaling factor
        sf1 = np.abs(vfuncXEv1[t4c[i, idxNeg1]]) / (vfuncXEv1[t4c[i, idxPos1]] - vfuncXEv1[t4c[i, idxNeg1]])

        # difference vector
        dv1 = v4c[t4c[i, idxPos1], :] - v4c[t4c[i, idxNeg1], :]

        # compute new point
        pt1 = sf1 * dv1 + v4c[t4c[i, idxNeg1], :]

        # --------------------------------------------------------------
        # store and connect

        # check for duplicate points (pt1)
        if newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]] == 0:
            # add to list
            v4c = np.concatenate((v4c, np.array(pt1, ndmin=2)))
            # keep index
            idxPt1 = len(v4c) - 1
            # update adj mtx
            newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]] = idxPt1
            newVtcsAdj[t4c[i, idxNeg1], t4c[i, idxPos1]] = idxPt1
            # also append i4c and k4c
            i4c = np.concatenate((i4c, st.mode(i4c[t4c[i, :]])[0]))
            k4c = np.concatenate((k4c, st.mode(k4c[t4c[i, :]])[0]))
            # let's remember the new indices, and whether or not it's a
            # 'positive' or 'negative' tetra (in terms of the 2nd
            # eigenfunction)
            newVtcs = np.concatenate((newVtcs, [idxPt1]))
            tmpSgn = np.unique(vfuncXEv2[t4c[i, :]][np.logical_not(np.isnan(vfuncXEv2[t4c[i, :]]))])
            if len(tmpSgn) == 1:
                newVtcsSgn = np.concatenate((newVtcsSgn, tmpSgn))
            else:
                newVtcsSgn = np.concatenate((newVtcsSgn, np.array([0])))
                raise RuntimeError("Inconsistency while creating seam 5, exiting.")
        else:
            idxPt1 = newVtcsAdj[t4c[i, idxPos1], t4c[i, idxNeg1]]

        # create new tetras
        # pos1, pt1, nan1, nan2
        # neg1, pt1, nan1, nan2
        newTetra = np.concatenate(
            (newTetra, np.array((t4c[i, idxPos1], idxPt1, t4c[i, idxNan1], t4c[i, idxNan2]), ndmin=2))
        )
        newTetra = np.concatenate(
            (newTetra, np.array((t4c[i, idxNeg1], idxPt1, t4c[i, idxNan1], t4c[i, idxNan2]), ndmin=2))
        )

        # return
        return v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn

    # --------------------------------------------------------------------------
    # getSeam subfunction #3

    # case 3: zero nans in the tetra (could happen for boundary tetras
    # towards two sides)
    def getSeamCase3(v4c, t4c, i4c, k4c, vfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i):
        # determine which edges of the current tetra belong to the
        # boundary surface
        numEdgeBnd = np.where(
            (
                np.logical_or(t4c[i, 0] == np.unique(e4cBndOpen, axis=0), t4c[i, 1] == np.unique(e4cBndOpen, axis=0))
                .all(axis=1)
                .any(),
                np.logical_or(t4c[i, 0] == np.unique(e4cBndOpen, axis=0), t4c[i, 2] == np.unique(e4cBndOpen, axis=0))
                .all(axis=1)
                .any(),
                np.logical_or(t4c[i, 0] == np.unique(e4cBndOpen, axis=0), t4c[i, 3] == np.unique(e4cBndOpen, axis=0))
                .all(axis=1)
                .any(),
                np.logical_or(t4c[i, 1] == np.unique(e4cBndOpen, axis=0), t4c[i, 2] == np.unique(e4cBndOpen, axis=0))
                .all(axis=1)
                .any(),
                np.logical_or(t4c[i, 1] == np.unique(e4cBndOpen, axis=0), t4c[i, 3] == np.unique(e4cBndOpen, axis=0))
                .all(axis=1)
                .any(),
                np.logical_or(t4c[i, 2] == np.unique(e4cBndOpen, axis=0), t4c[i, 3] == np.unique(e4cBndOpen, axis=0))
                .all(axis=1)
                .any(),
            )
        )[0]

        # number of trias that belong to the surface (and which ones)
        numTriaBnd = np.where(
            (
                np.logical_or(np.logical_or(t4c[i, 0] == t4cBndOpen, t4c[i, 1] == t4cBndOpen), t4c[i, 2] == t4cBndOpen)
                .all(axis=1)
                .any(),
                np.logical_or(np.logical_or(t4c[i, 0] == t4cBndOpen, t4c[i, 1] == t4cBndOpen), t4c[i, 3] == t4cBndOpen)
                .all(axis=1)
                .any(),
                np.logical_or(np.logical_or(t4c[i, 0] == t4cBndOpen, t4c[i, 2] == t4cBndOpen), t4c[i, 3] == t4cBndOpen)
                .all(axis=1)
                .any(),
                np.logical_or(np.logical_or(t4c[i, 1] == t4cBndOpen, t4c[i, 2] == t4cBndOpen), t4c[i, 3] == t4cBndOpen)
                .all(axis=1)
                .any(),
            )
        )[0]

        # examine values: we can only cut trias or edges that have at
        # least one pos and one neg value

        # number of edges that have pos / neg values (and which ones)
        numEdgeCross = np.where(
            np.sum(np.sign(vfuncXEv1[t4c[i, ((0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3))]]), axis=1) == 0
        )[0]

        # number of trias that have pos / neg values (and which ones)
        numTriaCross = np.where(
            np.abs(np.sum(np.sign(vfuncXEv1[t4c[i, ((0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3))]]), axis=1)) == 1
        )[0]

        #
        # print(
        #    "surface edges: " + str(len(numEdgeBnd)) + ", " +
        #    "surface trias: " + str(len(numTriaBnd)) + ", " +
        #    "pos/neg edges: " + str(len(numEdgeCross)) + ", " +
        #    "pos/neg trias: " + str(len(numTriaCross)) + ", " +
        #    "pos/neg surface edges: " + str(len(np.intersect1d(numEdgeBnd, numEdgeCross))) + ", " +
        #    "pos/neg surface trias: " + str(len(np.intersect1d(numTriaBnd, numTriaCross))))

        # we now analyze, based on the above info, which case the current one
        # actually is, i.e. ignoring the vertex or edge that is at the opposite
        # site etc.

        # distinguish cases
        if (
            len(numEdgeBnd) == 2
            and len(numTriaBnd) == 0
            and len(numEdgeCross) == 3
            and len(numTriaCross) == 3
            and len(np.intersect1d(numEdgeBnd, numEdgeCross)) == 1
            and len(np.intersect1d(numTriaBnd, numTriaCross)) == 0
        ):
            # surface edges: 2, surface trias: 0, pos/neg edges: 3, pos/neg trias: 3, pos/neg surface edges: 1, pos/neg surface trias: 0
            # should be case 2: two nans in tetra, due to only one single pos/neg surface edge
            # now find which edge is on the surface, and set the other points
            # temporarily to nan
            tmpvfuncXEv1 = vfuncXEv1.copy()
            tmpvfuncXEv1[
                t4c[
                    i,
                    np.setdiff1d(
                        (0, 1, 2, 3),
                        ((0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3))[
                            np.intersect1d(numEdgeBnd, numEdgeCross).item()
                        ],
                    ),
                ]
            ] = np.nan
            v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn = getSeamCase2(
                v4c, t4c, i4c, k4c, tmpvfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i
            )
        elif (
            len(numEdgeBnd) == 3
            and len(numTriaBnd) == 1
            and len(numEdgeCross) == 3
            and len(numTriaCross) == 3
            and len(np.intersect1d(numEdgeBnd, numEdgeCross)) == 2
            and len(np.intersect1d(numTriaBnd, numTriaCross)) == 1
        ):
            # surface edges: 3, surface trias: 1, pos/neg edges: 3, pos/neg trias: 3, pos/neg surface edges: 2, pos/neg surface trias: 1
            # should be case 1: one nan in tetra, due to one single surface tria
            # now find which tria is on the surface, and set the remaining vtx
            # temporarily to nan
            tmpvfuncXEv1 = vfuncXEv1.copy()
            tmpvfuncXEv1[
                t4c[
                    i,
                    np.setdiff1d(
                        (0, 1, 2, 3),
                        ((0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3))[np.intersect1d(numTriaBnd, numTriaCross).item()],
                    ),
                ]
            ] = np.nan
            v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn = getSeamCase1(
                v4c, t4c, i4c, k4c, tmpvfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i
            )
        elif (
            len(numEdgeBnd) == 3
            and len(numTriaBnd) == 1
            and len(numEdgeCross) == 4
            and len(numTriaCross) == 4
            and len(np.intersect1d(numEdgeBnd, numEdgeCross)) == 2
            and len(np.intersect1d(numTriaBnd, numTriaCross)) == 1
        ):
            # surface edges: 3, surface trias: 1, pos/neg edges: 4, pos/neg trias: 4, pos/neg surface edges: 2, pos/neg surface trias: 1
            # should be case 1: one nan in tetra, due to one single surface tria
            # now find which tria is on the surface, and set the remaining vtx
            # temporarily to nan - essentially the same case as above, but kept
            # separate for clarity
            tmpvfuncXEv1 = vfuncXEv1.copy()
            tmpvfuncXEv1[
                t4c[
                    i,
                    np.setdiff1d(
                        (0, 1, 2, 3),
                        ((0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3))[np.intersect1d(numTriaBnd, numTriaCross).item()],
                    ),
                ]
            ] = np.nan
            v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn = getSeamCase1(
                v4c, t4c, i4c, k4c, tmpvfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i
            )
        else:
            raise RuntimeError("Unknown getSeam() case, exiting.")

        # return
        return v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn

    # --------------------------------------------------------------------------
    # main part of function
    # --------------------------------------------------------------------------

    # initialize
    newVtcs = np.empty(shape=(0), dtype=int)
    newVtcsSgn = np.empty(shape=(0), dtype=int)
    newVtcsAdj = sp.lil_matrix((len(v4c), len(v4c)), dtype=int)
    newTetra = np.empty(shape=(0, 4), dtype=int)

    # assign first eigenfunction
    vfuncXEv1 = np.zeros(np.shape(v4c)[0]) * np.nan
    vfuncXEv1[v4cBndOpenKeep] = anisoLaplEvec[:, 1]

    # assign second eigenfunction
    vfuncXEv2 = np.zeros(np.shape(v4c)[0]) * np.nan
    vfuncXEv2[v4cBndOpenKeep] = np.sign(anisoLaplEvec[:, 2])

    # find tetra with zeros: first, transform boundary trias to edges; here,
    # the important trick is to start working with the boundary trias from
    # t4cBndOpen, which is to avoid trias that have e.g. 2 points at the upper
    # surface and 1 point on the lower surface (which can happen). Then
    # transform to edges and identify those with pos and neg values. We use
    # edges because these need to be cut also, not only trias, and because a
    # tetra can also just share an edge with the surface, not necessarily a
    # tria. Finally select all tetras that include one such edge.
    e4cBndOpen = np.concatenate((t4cBndOpen[:, (0, 1)], t4cBndOpen[:, (0, 2)], t4cBndOpen[:, (1, 2)]), axis=0)
    e4cBndOpen = np.unique(np.sort(e4cBndOpen, axis=1), axis=1)
    zeroEdgeIdx = e4cBndOpen[np.where(np.abs(np.sum(np.sign(vfuncXEv1[e4cBndOpen]), axis=1)) == 0)[0]]
    zeroTetraIdx = list()
    for i in zeroEdgeIdx:
        zeroTetraIdx.extend(np.where(np.sum(np.isin(t4c, i), axis=1) == 2)[0].tolist())
    zeroTetraIdx = np.unique(np.array(zeroTetraIdx))

    # from lapy import plot as lpp
    #
    # tetMesh = TetMesh(v4c, t4c)
    #
    # vfunc = np.zeros(np.shape(v4c)[0])
    # vfunc[np.unique(t4c[zeroTetraIdx])] = 1
    #
    # lpp.plot_tet_mesh(tetMesh, html_output=True, vfunc=vfunc, plot_edges=True, width=1200, height=1200)

    # loop across tetras with zeros (locally disable warning that np.nan < 0 is
    # False and np.nan > 0 is also False, which is intended)
    with np.errstate(invalid="ignore"):
        for i in zeroTetraIdx:
            # we can typically have one or two nans in the tetra. we may
            # sometimes have zero nans in the tetra.
            if np.sum(np.isnan(vfuncXEv1[t4c[i]])) == 1:
                # if there is one nan, we need to make sure that the three vtcs form a surface triangle; alternatively we check if it is at least a surface edge
                if np.isin(t4cBndOpen, t4c[i, np.logical_not(np.isnan(vfuncXEv1[t4c[i]]))]).all(axis=1).any():
                    v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn = getSeamCase1(
                        v4c, t4c, i4c, k4c, vfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i
                    )
                elif np.isin(e4cBndOpen, t4c[i, np.logical_not(np.isnan(vfuncXEv1[t4c[i]]))]).all(axis=1).any():
                    # need to temporarily set one vtx to nan
                    tmpEdge = (
                        np.isin(e4cBndOpen, t4c[i, (0, 1)]).all(axis=1).any(),
                        np.isin(e4cBndOpen, t4c[i, (0, 2)]).all(axis=1).any(),
                        np.isin(e4cBndOpen, t4c[i, (0, 3)]).all(axis=1).any(),
                        np.isin(e4cBndOpen, t4c[i, (1, 2)]).all(axis=1).any(),
                        np.isin(e4cBndOpen, t4c[i, (1, 3)]).all(axis=1).any(),
                        np.isin(e4cBndOpen, t4c[i, (2, 3)]).all(axis=1).any(),
                    )
                    if len(np.where(tmpEdge)) > 1:
                        raise RuntimeError("Found " + str(len(np.where(tmpEdge))) + " edges, exiting.")
                    tmpEdgeNot = np.setdiff1d(
                        (0, 1, 2, 3), ((0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3))[np.where(tmpEdge)[0].item()]
                    )
                    tmpvfuncXEv1 = vfuncXEv1.copy()
                    tmpvfuncXEv1[t4c[i, tmpEdgeNot]] = np.nan
                    v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn = getSeamCase2(
                        v4c, t4c, i4c, k4c, tmpvfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i
                    )
                else:
                    raise RuntimeError("Vtx " + str(i) + " not in triangle or edge, skipping")

            elif np.sum(np.isnan(vfuncXEv1[t4c[i]])) == 2:
                # if there are two nans, we need to make sure that they are linked by a surface edge; this should have happenend already.
                tmpEdge = (
                    np.isin(e4cBndOpen, t4c[i, (0, 1)]).all(axis=1).any(),
                    np.isin(e4cBndOpen, t4c[i, (0, 2)]).all(axis=1).any(),
                    np.isin(e4cBndOpen, t4c[i, (0, 3)]).all(axis=1).any(),
                    np.isin(e4cBndOpen, t4c[i, (1, 2)]).all(axis=1).any(),
                    np.isin(e4cBndOpen, t4c[i, (1, 3)]).all(axis=1).any(),
                    np.isin(e4cBndOpen, t4c[i, (2, 3)]).all(axis=1).any(),
                )
                if len(np.where(tmpEdge)) > 1:
                    raise RuntimeError("Found " + str(len(np.where(tmpEdge))) + " edges, exiting.")
                v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn = getSeamCase2(
                    v4c, t4c, i4c, k4c, vfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i
                )

            elif np.sum(np.isnan(vfuncXEv1[t4c[i]])) == 0:
                # if there are no nans, the checking is done within the getSeamCase3 function
                v4c, t4c, i4c, k4c, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn = getSeamCase3(
                    v4c, t4c, i4c, k4c, vfuncXEv1, e4cBndOpen, newTetra, newVtcs, newVtcsAdj, newVtcsSgn, i
                )

            else:
                # we should never have four nans
                raise RuntimeError("Inconsistency while creating seam (incorrect number of nans), exiting.")

    # remove tetras
    t4c = np.delete(t4c, zeroTetraIdx, axis=0)

    # add new tetras
    t4c = np.concatenate((t4c, newTetra))

    # return
    return v4c, t4c, i4c, k4c, newVtcs, newVtcsSgn, newTetra


# ------------------------------------------------------------------------------
# MAIN FUNCTION
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# computeCubeParam


def computeCubeParam(params):
    # --------------------------------------------------------------------------
    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Computing cube parametrization")
    print()

    # --------------------------------------------------------------------------
    # get data

    # evaluate params

    cutTetraMeshDir = os.path.join(params.OUTDIR, "tetra-cube")
    cutTetraMeshFile = os.path.join(params.OUTDIR, params.HEMI + ".cut.vtk")
    cutTetraIndexFile = os.path.join(params.OUTDIR, "tetra-cut", params.HEMI + ".cut.psol")
    openBndCutTetraMeshFile = os.path.join(params.OUTDIR, "tetra-cut", params.HEMI + ".open.bnd.cut.vtk")
    tetraIndexFile = os.path.join(params.OUTDIR, "tetra-labels", params.HEMI + ".tetra.psol")

    if len(params.internal.aniso_alpha) == 1:
        paramAnisoAlpha = params.internal.aniso_alpha[0]
    else:
        paramAnisoAlpha = params.internal.aniso_alpha

    # load cut tetra mesh

    tetMeshCut = TetMesh.read_vtk(cutTetraMeshFile)
    v4c = tetMeshCut.v
    t4c = tetMeshCut.t

    # load indices of cutting planes

    i4c = io.read_vfunc(cutTetraIndexFile)
    i4c = np.array(i4c)

    # load subfields indices

    j4c = io.read_vfunc(tetraIndexFile)
    j4c = np.array(j4c)

    # append subfield indices

    k4c = np.concatenate((j4c, np.zeros(len(i4c) - len(j4c))))

    # --------------------------------------------------------------------------
    # get a tria mesh that is cut open at its ends, but still consistent with
    # the tetra mesh

    triaMesh4cBndOpen = TriaMesh.read_vtk(openBndCutTetraMeshFile)

    v4cBndOpen = triaMesh4cBndOpen.v
    t4cBndOpen = triaMesh4cBndOpen.t

    # remove free vertices and orient

    v4cBndOpenKeep, v4cBndOpenDel = triaMesh4cBndOpen.rm_free_vertices_()

    triaMesh4cBndOpenRm = TriaMesh(v=triaMesh4cBndOpen.v, t=triaMesh4cBndOpen.t)

    triaMesh4cBndOpenRm.orient_()

    v4cBndOpenRm = triaMesh4cBndOpenRm.v
    t4cBndOpenRm = triaMesh4cBndOpenRm.t

    # --------------------------------------------------------------------------
    # compute anisotropic laplace

    # compute first eigenfunction

    fem = Solver(triaMesh4cBndOpenRm, lump=True, aniso=paramAnisoAlpha, aniso_smooth=params.internal.aniso_smooth)

    anisoLaplEval, anisoLaplEvec = fem.eigs(k=3)

    # --------------------------------------------------------------------------
    # get mapping of open boundary vertices to subfields

    with open(os.path.join(os.path.dirname(openBndCutTetraMeshFile), params.HEMI + ".rm.open.bnd.cut.lst"), "r") as f:
        hsfList = f.read().splitlines()

    hsfList = np.array(hsfList).astype(int)

    io.write_vfunc(
        os.path.join(os.path.dirname(openBndCutTetraMeshFile), params.HEMI + ".hsf.rm.open.bnd.cut.psol"), k4c[hsfList]
    )

    # --------------------------------------------------------------------------
    # post-process eigenfunction (order, flipping)

    # decide whether or not to flip anisoLaplEvec[:, 1] and anisoLaplEvec[:, 2]:
    # - anisoLaplEvec[:, 1] should be zero at the extrema of 234 and 240.
    # - anisoLaplEvec[:, 2] should have extremal values at 234 and 240.
    # We therefore check if the zeros of Evec1 are located in 234 or 240/2420
    # rather than 236 or 238. If not, we change order.

    # assign first eigenfunction
    vfuncXEv1 = np.zeros(np.shape(v4c)[0]) * np.nan
    vfuncXEv1[v4cBndOpenKeep] = anisoLaplEvec[:, 1]
    # get edges that contain zeros, and turn into list of vertices
    e4cBndOpen = np.concatenate((t4cBndOpen[:, (0, 1)], t4cBndOpen[:, (0, 2)], t4cBndOpen[:, (1, 2)]), axis=0)
    e4cBndOpen = np.unique(np.sort(e4cBndOpen, axis=1), axis=1)
    zeroEdgeIdx = e4cBndOpen[np.where(np.abs(np.sum(np.sign(vfuncXEv1[e4cBndOpen]), axis=1)) == 0)[0]]

    # decide whether or not to flip anisoLaplEvec[:, 1]  (should be inf -> sup)

    # check EV1 for flips; this is done indirectly: we get the locations of
    # highest and lowest values in 234 (PrSbc) or 236 (Sbc); highest values
    # should have a higher value on the z-axis than lower values.

    if np.median(
        v4cBndOpenRm[
            np.where(
                np.logical_and(
                    anisoLaplEvec[:, 1] > 0,
                    np.logical_or(
                        k4c[hsfList] == params.LUTDICT["presubiculum"], k4c[hsfList] == params.LUTDICT["subiculum"]
                    ),
                )
            )[0],
            2,
        ]
    ) > np.median(
        v4cBndOpenRm[
            np.where(
                np.logical_and(
                    anisoLaplEvec[:, 1] < 0,
                    np.logical_or(
                        k4c[hsfList] == params.LUTDICT["presubiculum"], k4c[hsfList] == params.LUTDICT["subiculum"]
                    ),
                )
            )[0],
            2,
        ]
    ):
        pass
    elif np.median(
        v4cBndOpenRm[
            np.where(
                np.logical_and(
                    anisoLaplEvec[:, 1] > 0,
                    np.logical_or(
                        k4c[hsfList] == params.LUTDICT["presubiculum"], k4c[hsfList] == params.LUTDICT["subiculum"]
                    ),
                )
            )[0],
            2,
        ]
    ) < np.median(
        v4cBndOpenRm[
            np.where(
                np.logical_and(
                    anisoLaplEvec[:, 1] < 0,
                    np.logical_or(
                        k4c[hsfList] == params.LUTDICT["presubiculum"], k4c[hsfList] == params.LUTDICT["subiculum"]
                    ),
                )
            )[0],
            2,
        ]
    ):
        LOGGER.info("Flipping EV1")
        anisoLaplEvec[:, 1] = -anisoLaplEvec[:, 1]
    else:
        io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".lapy.aLBO.EV1.psol"), anisoLaplEvec[:, 1])
        io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".lapy.aLBO.EV2.psol"), anisoLaplEvec[:, 2])
        raise RuntimeError("Inconsistency detected for EV1, exiting.")

    # decide whether or not to flip anisoLaplEvec[:, 2] (should be 234 -> 240)

    if (
        np.median(
            anisoLaplEvec[
                np.where(
                    np.logical_or(
                        k4c[hsfList] == params.LUTDICT["presubiculum"], k4c[hsfList] == params.LUTDICT["subiculum"]
                    )
                )[0],
                2,
            ]
        )
        < 0
    ):
        pass
    elif (
        np.median(
            anisoLaplEvec[
                np.where(
                    np.logical_or(
                        k4c[hsfList] == params.LUTDICT["presubiculum"], k4c[hsfList] == params.LUTDICT["subiculum"]
                    )
                )[0],
                2,
            ]
        )
        > 0
    ):
        LOGGER.info("Flipping EV2")
        anisoLaplEvec[:, 2] = -anisoLaplEvec[:, 2]
    else:
        io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".lapy.aLBO.EV1.psol"), anisoLaplEvec[:, 1])
        io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".lapy.aLBO.EV2.psol"), anisoLaplEvec[:, 2])
        raise RuntimeError("Inconsistency detected for EV2, exiting.")

    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".lapy.aLBO.EV1.psol"), anisoLaplEvec[:, 1])
    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".lapy.aLBO.EV2.psol"), anisoLaplEvec[:, 2])

    # --------------------------------------------------------------------------
    # define boundary conditions (note that we changed dimensions and directions!)

    # get seam

    v4c, t4c, i4c, k4c, newVtcs, newVtcsSgn, newTetra = getSeam(
        v4c, t4c, i4c, k4c, v4cBndOpenKeep, t4cBndOpen, anisoLaplEvec
    )

    # get 2nd principal axis, note that direction may be arbitrary, will
    # possibly be switched below; note that this replaces EV2-based newVtcsSgn,
    # which we may remove in the future

    pca = PCA(n_components=3)
    pca_scores = pca.fit_transform(v4c)

    # x: medial -> lateral

    vfuncX = np.zeros(np.shape(v4c)[0])
    srcX = np.zeros(np.shape(v4c)[0])
    srcX[newVtcs[v4c[newVtcs, 0] < np.mean(v4c[newVtcs, 0])]] = -1
    ###srcX[newVtcs[pca_scores[newVtcs,1]<0]] = -1
    ###srcX[newVtcs[v4c[newVtcs,0]>0]] = 1
    ###srcX[newVtcs[newVtcsSgn>0]] = 1
    snkX = np.zeros(np.shape(v4c)[0])
    snkX[newVtcs[v4c[newVtcs, 0] >= np.mean(v4c[newVtcs, 0])]] = 1
    ###snkX[newVtcs[pca_scores[newVtcs,1]>0]] = 1
    ###snkX[newVtcs[v4c[newVtcs,0]<0]] = -1
    ###snkX[newVtcs[newVtcsSgn<0]] = -1
    vfuncX = srcX + snkX

    # y: Tail (226) -> Head (232) (post --> ant)

    vfuncY = np.zeros(np.shape(v4c)[0])
    vfuncY[np.where(i4c == params.LUTDICT["jointtail"])[0]] = -1
    vfuncY[np.where(i4c == params.LUTDICT["jointhead"])[0]] = 1

    # z: inf->sup

    vfuncZ = np.zeros(np.shape(v4c)[0])
    srcZ = np.zeros(np.shape(anisoLaplEvec[:, 1])[0])
    srcZ[anisoLaplEvec[:, 1] < 0] = -1
    snkZ = np.zeros(np.shape(anisoLaplEvec[:, 1])[0])
    snkZ[anisoLaplEvec[:, 1] > 0] = 1
    vfuncZ[v4cBndOpenKeep] = srcZ + snkZ

    # determine left / right

    img = nb.load(os.path.join(params.OUTDIR, params.HEMI + ".image.mgz"))
    mat = np.linalg.inv(img.header.get_vox2ras_tkr())
    pts = np.concatenate((v4c, np.ones((v4c.shape[0], 1))), axis=1)
    ras = np.matmul(np.matmul(pts, mat.transpose()), img.header.get_vox2ras().transpose())

    if params.HEMI == "lh":
        if (ras[:, 0] < 0.0).all():
            # lateral (snk, +1) should be more negative than medial (src, -1), otherwise flip
            if np.mean(ras[np.where(vfuncX > 0)[0], 0]) > np.mean(ras[np.where(vfuncX < 0)[0], 0]):
                vfuncX = -vfuncX
        else:
            raise RuntimeError("Ambiguous hemisphere info, exiting.")
    elif params.HEMI == "rh":
        if (ras[:, 0] > 0.0).all():
            # lateral (snk, +1) should be more positive than medial (src, -1), otherwise flip
            if np.mean(ras[np.where(vfuncX > 0)[0], 0]) < np.mean(ras[np.where(vfuncX < 0)[0], 0]):
                vfuncX = -vfuncX
        else:
            raise RuntimeError("Ambiguous hemisphere info, exiting.")

    # --------------------------------------------------------------------------
    # compute ABtetra and Poisson functions

    # first, need to remove free vtcs

    tetMesh4c = TetMesh(v4c, t4c)

    v4cRmKeep, v4cRmDel = tetMesh4c.rm_free_vertices_()

    v4cRm = tetMesh4c.v
    t4cRm = tetMesh4c.t

    TetMesh.write_vtk(tetMesh4c, os.path.join(cutTetraMeshDir, params.HEMI + ".seam.rm.cut.vtk"))

    #

    vfuncXRm = vfuncX[v4cRmKeep]
    vfuncYRm = vfuncY[v4cRmKeep]
    vfuncZRm = vfuncZ[v4cRmKeep]

    #

    bndXRm = (np.where(vfuncXRm)[0], vfuncXRm[np.where(vfuncXRm)])
    bndYRm = (np.where(vfuncYRm)[0], vfuncYRm[np.where(vfuncYRm)])
    bndZRm = (np.where(vfuncZRm)[0], vfuncZRm[np.where(vfuncZRm)])

    #

    tetMesh4cRm = TetMesh(v4cRm, t4cRm)

    fem = Solver(tetMesh4cRm)

    P0 = fem.poisson(dtup=bndXRm)
    P1 = fem.poisson(dtup=bndYRm)
    P2 = fem.poisson(dtup=bndZRm)

    # --------------------------------------------------------------------------
    # output

    # write out functions

    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".vfuncX.seam.rm.cut.psol"), vfuncXRm)
    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".vfuncY.seam.rm.cut.psol"), vfuncYRm)
    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".vfuncZ.seam.rm.cut.psol"), vfuncZRm)

    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".poisson0.seam.rm.cut.psol"), P0)
    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".poisson1.seam.rm.cut.psol"), P1)
    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".poisson2.seam.rm.cut.psol"), P2)

    # also write out boundary files for visualization

    tetMesh4cRmBnd = tetMesh4cRm.boundary_tria()

    v4cRmBndRmKeep, v4cRmBndRmDel = tetMesh4cRmBnd.rm_free_vertices_()

    v4cRmBndRm = tetMesh4cRmBnd.v
    t4cRmBndRm = tetMesh4cRmBnd.t

    TriaMesh.write_vtk(tetMesh4cRmBnd, filename=os.path.join(cutTetraMeshDir, params.HEMI + ".rm.bnd.seam.rm.cut.vtk"))

    io.write_vfunc(
        os.path.join(cutTetraMeshDir, params.HEMI + ".vfuncX.rm.bnd.seam.rm.cut.psol"), vfuncXRm[v4cRmBndRmKeep]
    )
    io.write_vfunc(
        os.path.join(cutTetraMeshDir, params.HEMI + ".vfuncY.rm.bnd.seam.rm.cut.psol"), vfuncYRm[v4cRmBndRmKeep]
    )
    io.write_vfunc(
        os.path.join(cutTetraMeshDir, params.HEMI + ".vfuncZ.rm.bnd.seam.rm.cut.psol"), vfuncZRm[v4cRmBndRmKeep]
    )

    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".poisson0.rm.bnd.seam.rm.cut.psol"), P0[v4cRmBndRmKeep])
    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".poisson1.rm.bnd.seam.rm.cut.psol"), P1[v4cRmBndRmKeep])
    io.write_vfunc(os.path.join(cutTetraMeshDir, params.HEMI + ".poisson2.rm.bnd.seam.rm.cut.psol"), P2[v4cRmBndRmKeep])

    # write out cube

    w4cRm = np.vstack((P0, P1, P2)).transpose()

    TetMesh.write_vtk(TetMesh(v=w4cRm, t=t4cRm), os.path.join(cutTetraMeshDir, params.HEMI + ".uvw.seam.rm.cut.vtk"))

    #

    return params



"""
This module provides configuration defaults for the hippocampal shape and 
thickness analysis package.

"""


def get_defaults(x):
    defaults = dict(
        # note: if changing these values, remember to adjust the argparse help
        # messages as well as the run_hipsta docstring also
        # options
        no_cleanup=False,
        no_crop=False,
        upsample=False,
        upsample_size=[0, 0, 0],
        no_merge_molecular_layer=False,
        automask_head=False,
        automask_tail=False,
        automask_head_margin=0,
        automask_tail_margin=0,
        no_gauss_filter=False,
        gauss_filter_size=[1, 50],
        long_filter=False,
        long_filter_size=5,
        no_close_mask=False,
        mca="mri_mc",
        remesh=False,
        smooth=5,
        cut_range=[-0.975, 0.975],
        aniso_alpha=[40],
        aniso_smooth=3,
        thickness_grid=[-0.9, 0.9, 41, -0.975, 0.975, 21, -0.9, 0.9, 11],
        # expert options
        mcc=1,
        remesh_size=0,
        no_check_surface=False,
        no_check_boundaries=False,
        no_qc=False,
        allow_ragged_surfaces=False,
        allow_ragged_trias=False,
        # deprecated options
        no_orient=False,
        # internal options (not set during parsing or class definition, but during args evaluation)
        map_values_integrate="none",
        map_values_select=None,
        map_values_interp="nearest",
        map_values_write_psol=True,
        map_values_write_mgh=True,
        map_values_write_annot=False,
    )

    if x in defaults.keys():
        return defaults[x]
    else:
        raise ValueError(x + " is not a valid key in the defaults dict")



"""
This module provides a function to create a tetrahedral mesh

"""

import os
import shutil
import subprocess

from lapy import TetMesh, TriaMesh

# ==============================================================================
# FUNCTIONS

# ------------------------------------------------------------------------------
# _createSTL


def _createSTL(filename, v, t):
    """
    createSTL(filename, v, t)

    A function to write STL files

    """

    import numpy as np

    with open(filename, "w") as f:
        print("solid lap-data", file=f)

        for i in range(0, len(t)):
            v1mv0 = v[t[i, 1], :] - v[t[i, 0], :]
            v2mv0 = v[t[i, 2], :] - v[t[i, 0], :]

            normal = np.cross(v1mv0, v2mv0) / np.linalg.norm(np.cross(v1mv0, v2mv0))

            print("  facet normal", file=f)

            print("    outer loop", file=f)

            print("      vertex %f %f %f " % (v[t[i, 0], 0], v[t[i, 0], 1], v[t[i, 0], 2]), file=f)
            print("      vertex %f %f %f " % (v[t[i, 1], 0], v[t[i, 1], 1], v[t[i, 1], 2]), file=f)
            print("      vertex %f %f %f " % (v[t[i, 2], 0], v[t[i, 2], 1], v[t[i, 2], 2]), file=f)

            print("    endloop", file=f)

            print("  endfacet", file=f)

        print("end solid lap-data", file=f)


# ------------------------------------------------------------------------------
# createTetraMesh


def createTetraMesh(params):
    """ """

    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Create tetrahedral mesh")
    print()

    # export mesh as STL

    triaMesh = TriaMesh.read_vtk(os.path.join(params.OUTDIR, params.HEMI + ".surf.vtk"))

    _createSTL(os.path.join(params.OUTDIR, "tetra-mesh", params.HEMI + ".tetra.stl"), v=triaMesh.v, t=triaMesh.t)

    # create geofile (test.geo)

    with open(os.path.join(params.OUTDIR, "tetra-mesh", params.HEMI + ".tetra.geo"), "w") as f:
        print("Mesh.Algorithm3D=4;", file=f)
        print("Mesh.Optimize=1;", file=f)
        print("Mesh.OptimizeNetgen=1;", file=f)
        print('Merge "' + params.HEMI + ".tetra.stl" + '";', file=f)
        print("Surface Loop(1) = {1};", file=f)
        print("Volume(1) = {1};", file=f)
        print("Physical Volume(1) = {1};", file=f)

    # use gmsh to get tetras

    cmd = (
        shutil.which("gmsh")
        + " "
        + "-3 "
        + "-o "
        + os.path.join(params.OUTDIR, "tetra-mesh", params.HEMI + ".tetra-gmsh.vtk")
        + " "
        + os.path.join(params.OUTDIR, "tetra-mesh", params.HEMI + ".tetra.geo")
    )

    print(cmd)

    subprocess.run(cmd.split(), capture_output=True)

    # convert vtk2 to vtk1

    tetMesh = TetMesh.read_vtk(os.path.join(params.OUTDIR, "tetra-mesh", params.HEMI + ".tetra-gmsh.vtk"))

    TetMesh.write_vtk(tetMesh, os.path.join(params.OUTDIR, params.HEMI + ".tetra.vtk"))

    # return

    return params



"""
hipsta CLI
"""

from ..hipsta import _parse_arguments, _run_hipsta, get_help


def main():
    # message

    print("")
    print("----------------------------------------")
    print("Hippocampal shape and thickness analysis")
    print("----------------------------------------")
    print("")

    # parse arguments

    args = _parse_arguments()

    # return help or run analysis

    if args.more_help is True:
        get_help()
    else:
        _run_hipsta(args)



"""
This module provides functions for image processing operations (such as binarization, filtering, closing operations)

"""

import os
import shutil

import nibabel as nb
import numpy as np
from scipy import ndimage

# ==============================================================================
# FUNCTIONS


def gaussFilter(params):
    """ """

    #

    if params.internal.GAUSSFILTER is True:
        # message

        print()
        print("--------------------------------------------------------------------------------")
        print("Gaussian filtering")
        print()

        # gaussian filtering

        img = nb.load(params.FILENAME)
        dat = img.get_fdata()

        dat_filtered = (
            ndimage.gaussian_filter(
                (dat > 0) * 100, sigma=params.internal.GAUSSFILTER_SIZE[0], mode="constant", cval=0.0
            )
            > params.internal.GAUSSFILTER_SIZE[1]
        )

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=dat_filtered.astype("float32"), affine=img.affine),
            filename=os.path.join(params.OUTDIR, "mask", params.HEMI + ".gaussian_filter.mgz"),
        )

        # update params

        params.FILENAME = os.path.join(params.OUTDIR, "mask", params.HEMI + ".gaussian_filter.mgz")

    # return

    return params


def longFilter(params):
    """ """

    #

    if params.internal.LONGFILTER is True:
        # message

        print()
        print("--------------------------------------------------------------------------------")
        print("Filtering along longitudinal axis")
        print()

        # longitudinal filtering

        img = nb.load(params.FILENAME)
        dat = img.get_fdata()

        k = np.zeros(
            (params.internal.LONGFILTER_SIZE, params.internal.LONGFILTER_SIZE, params.internal.LONGFILTER_SIZE)
        )
        k[int((params.internal.LONGFILTER_SIZE - 1) / 2), int((params.internal.LONGFILTER_SIZE - 1) / 2), :] = 1

        dat_filtered = ndimage.convolve(dat, k, mode="constant", cval=0.0)

        dat_filtered = dat_filtered > 0

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=dat_filtered.astype("float32"), affine=img.affine),
            filename=os.path.join(params.OUTDIR, "mask", params.HEMI + ".longitudinal_filter.mgz"),
        )

        # update params

        params.FILENAME = os.path.join(params.OUTDIR, "mask", params.HEMI + ".longitudinal_filter.mgz")

    # return

    return params


def closeMask(params):
    """ """

    #

    if params.internal.CLOSEMASK is True:
        # message

        print()
        print("--------------------------------------------------------------------------------")
        print("Applying closing operation to mask")
        print()

        # get data

        img = nb.load(params.FILENAME)
        dat = img.get_fdata()

        dat_filtered = ndimage.binary_closing(
            dat,
            structure=ndimage.iterate_structure(
                ndimage.generate_binary_structure(rank=3, connectivity=1), iterations=1
            ).astype(int),
            iterations=1,
        )

        dat_filtered = dat_filtered > 0

        nb.freesurfer.save(
            nb.freesurfer.MGHImage(dataobj=dat_filtered.astype("float32"), affine=img.affine),
            filename=os.path.join(params.OUTDIR, "mask", params.HEMI + ".close_mask.mgz"),
        )

        # update params

        params.FILENAME = os.path.join(params.OUTDIR, "mask", params.HEMI + ".close_mask.mgz")

    # return

    return params


def binarizeMask(params):
    # binarize

    img = nb.load(params.FILENAME)
    dat = img.get_fdata()

    dat_filtered = dat != 0

    nb.freesurfer.save(
        nb.freesurfer.MGHImage(dataobj=dat_filtered.astype("float32"), affine=img.affine),
        filename=os.path.join(params.OUTDIR, "mask", params.HEMI + ".initial_mask.mgz"),
    )

    # update params

    params.FILENAME = os.path.join(params.OUTDIR, "mask", params.HEMI + ".initial_mask.mgz")

    # return

    return params


def copy_mask_to_main(params):
    """ """

    # copy to main directory

    shutil.copyfile(params.FILENAME, os.path.join(params.OUTDIR, params.HEMI + ".mask.mgz"))

    # update params

    params.FILENAME = os.path.join(params.OUTDIR, params.HEMI + ".mask.mgz")

    # return

    return params



"""
This module provides a function to create surfaces, which includes the marching
cube algorithm, remeshing and smoothing.

"""

import os
import shutil
import subprocess

import nibabel as nb
import numpy as np
import pyacvd
import pyvista as pv
from lapy import TriaMesh
from scipy import sparse as sp
from skimage import measure as skm

# ==============================================================================
# FUNCTIONS


def extractSurface(params):
    """ """

    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Creating surface via marching cube algorithm")
    print()

    # create surface via marching cube algorithm

    if params.internal.MCA == "mri_mc":
        cmd = (
            os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_mc")
            + " "
            + os.path.join(params.OUTDIR, params.HEMI + ".mask.mgz")
            + " 1 "
            + os.path.join(params.OUTDIR, "surface", params.HEMI + ".initial_surf.vtk")
            + " "
            + str(params.internal.MCC)
        )

        print(cmd)

        subprocess.run(cmd.split(), capture_output=True)

    elif params.internal.MCA == "mri_tessellate":
        cmd = (
            os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_pretess")
            + " "
            + os.path.join(params.OUTDIR, params.HEMI + ".mask.mgz")
            + " xyz "
            + os.path.join(params.OUTDIR, params.HEMI + ".mask.mgz")
            + " "
            + os.path.join(params.OUTDIR, "surface", params.HEMI + ".pretess.mgz")
        )

        print(cmd)

        subprocess.run(cmd.split(), capture_output=True)

        cmd = (
            os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_tessellate")
            + " "
            + os.path.join(params.OUTDIR, "surface", params.HEMI + ".pretess.mgz")
            + " 1 "
            + os.path.join(params.OUTDIR, "surface", params.HEMI + ".surf.vtk")
        )

        print(cmd)

        subprocess.run(cmd.split(), capture_output=True)

        # convert from freesurfer binary surface format to vtk

        cmd = (
            os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mris_convert")
            + " "
            + os.path.join(params.OUTDIR, "surface", params.HEMI + ".surf.vtk")
            + " "
            + os.path.join(params.OUTDIR, "surface", params.HEMI + ".initial_surf.vtk")
        )

        print(cmd)

        subprocess.run(cmd.split(), capture_output=True)

    elif params.internal.MCA == "skimage":
        img = nb.load(os.path.join(params.OUTDIR, params.HEMI + ".mask.mgz"))
        dat = img.get_fdata()

        msh = skm.marching_cubes(dat, allow_degenerate=False, step_size=2)

        v = np.matmul(
            img.header.get_vox2ras_tkr(), np.concatenate((msh[0], np.ones((msh[0].shape[0], 1))), axis=1).T
        ).T[:, 0:3]
        t = msh[1]

        # keep largest component

        triaMesh = TriaMesh(v=v, t=t)
        comps = sp.csgraph.connected_components(triaMesh.adj_sym, directed=False)
        if comps[0] > 1:
            compsLargest = np.argmax(np.unique(comps[1], return_counts=True)[1])
            vtcsRemove = np.where(comps[1] != compsLargest)
            triaKeep = np.sum(np.isin(t, vtcsRemove), axis=1) == 0
            tcomp = t[triaKeep, :]
        else:
            tcomp = t

        # write to file

        tria_mesh = TriaMesh(v=v, t=tcomp)
        tria_mesh.rm_free_vertices_()
        TriaMesh.write_vtk(tria_mesh, os.path.join(params.OUTDIR, "surface", params.HEMI + ".initial_surf.vtk"))

    # update params

    params.SURFNAME = os.path.join(params.OUTDIR, "surface", params.HEMI + ".initial_surf.vtk")

    # return

    return params


def remeshSurface(params):
    if params.internal.REMESH is True:
        # message

        print()
        print("--------------------------------------------------------------------------------")
        print("Remesh surface")
        print()

        if shutil.which("mris_remesh") is None:
            Mesh = pv.PolyData(params.SURFNAME)

            clustered = pyacvd.Clustering(Mesh)
            clustered.subdivide(4)
            clustered.cluster(Mesh.n_points)

            remeshed = clustered.create_mesh()

            vr = remeshed.points
            tr = np.reshape(remeshed.faces, (int(len(remeshed.faces) / 4), 4))[:, 1:]

            # check remesh results

            # creates list of edges
            trSort = np.sort(tr, axis=1)
            trSortEdges = np.concatenate((trSort[:, [0, 1]], trSort[:, [1, 2]], trSort[:, [0, 2]]), axis=0)

            # remove trias that have an edge that only occurs once (boundary trias)
            countEdges = np.unique(trSortEdges, axis=0, return_counts=True)
            removeEdges = countEdges[0][np.where(countEdges[1] == 1)[0]]
            if len(removeEdges) > 0:
                removeTrias = np.unique(
                    [
                        np.where(np.sum(np.logical_or(trSort == i[0], trSort == i[1]), axis=1) == 2)[0]
                        for i in removeEdges
                    ]
                )
                trSortRmBnd = np.delete(trSort, removeTrias, axis=0)
            else:
                trSortRmBnd = trSort.copy()

            # assure that any edge occurs exactly two times (duplicates)
            trSortRmBndEdges = np.concatenate(
                (trSortRmBnd[:, [0, 1]], trSortRmBnd[:, [1, 2]], trSortRmBnd[:, [0, 2]]), axis=0
            )
            if len(np.where(np.unique(trSortRmBndEdges, axis=0, return_counts=True)[1] != 2)[0]) != 0:
                raise RuntimeError("Duplicate edges in mesh, exiting.")

            # assure that every edge must be part of exactly two different triangles (no boundary edges, no duplicates)
            countEdgesInTrias = np.array(
                [
                    np.sum(
                        np.sum(
                            np.logical_or(trSortRmBnd == trSortRmBndEdges[i, 0], trSortRmBnd == trSortRmBndEdges[i, 1]),
                            axis=1,
                        )
                        == 2
                    )
                    for i in range(0, len(trSortRmBndEdges))
                ]
            )
            if (countEdgesInTrias != 2).any():
                raise RuntimeError("Boundary or duplicate edges in mesh, exiting.")

            # restrict to largest connected component

            triaMesh = TriaMesh(v=vr, t=trSortRmBnd)
            comps = sp.csgraph.connected_components(triaMesh.adj_sym, directed=False)
            if comps[0] > 1:
                compsLargest = np.argmax(np.unique(comps[1], return_counts=True)[1])
                vtcsRemove = np.where(comps[1] != compsLargest)
                triaKeep = np.sum(np.isin(trSortRmBnd, vtcsRemove), axis=1) == 0
                trSortRmBndRmComps = trSortRmBnd[triaKeep, :]
            else:
                trSortRmBndRmComps = trSortRmBnd

            # remove free vertices and re-orient mesh

            triaMesh = TriaMesh(v=vr, t=trSortRmBndRmComps)
            triaMesh.rm_free_vertices_()
            triaMesh.orient_()

        else:
            if params.internal.REMESH_SIZE == 0:
                cmd = (
                    os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mris_remesh")
                    + " "
                    + "--remesh  -i "
                    + params.SURFNAME
                    + " "
                    + "-o "
                    + os.path.join(params.OUTDIR, "surface", params.HEMI + ".remeshed_surf.vtk")
                )
            elif params.internal.REMESH_SIZE > 0:
                cmd = (
                    os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mris_remesh")
                    + " "
                    + "--nvert "
                    + str(params.internal.REMESH_SIZE)
                    + "  -i "
                    + params.SURFNAME
                    + " "
                    + "-o "
                    + os.path.join(params.OUTDIR, "surface", params.HEMI + ".remeshed_surf.vtk")
                )

            print(cmd)

            subprocess.run(cmd.split(), capture_output=True)

            # remove free vertices and re-orient mesh

            triaMesh = TriaMesh.read_vtk(os.path.join(params.OUTDIR, "surface", params.HEMI + ".remeshed_surf.vtk"))
            triaMesh.rm_free_vertices_()
            triaMesh.orient_()

        # save

        TriaMesh.write_vtk(triaMesh, os.path.join(params.OUTDIR, "surface", params.HEMI + ".remeshed_surf.vtk"))

        # update params

        params.SURFNAME = os.path.join(params.OUTDIR, "surface", params.HEMI + ".remeshed_surf.vtk")

    # return

    return params


def smoothSurface(params):
    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Smooth surface")
    print()

    # read mesh

    triaMesh = TriaMesh.read_vtk(params.SURFNAME)

    # remove free vertices and re-orient mesh

    triaMesh.rm_free_vertices_()
    triaMesh.orient_()

    # smoothing

    triaMesh.smooth_(n=params.internal.SMO)

    # write mesh

    TriaMesh.write_vtk(triaMesh, os.path.join(params.OUTDIR, params.HEMI + ".surf.vtk"))

    # update params

    params.SURFNAME = os.path.join(params.OUTDIR, params.HEMI + ".surf.vtk")

    # return

    return params



"""
This module provides atlas information for the hippocampal shape and thickness 
analysis package.

"""

import logging
import os

import pandas

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS


def get_atlases(lut):
    if lut == "freesurfer":
        LOGGER.info("Found internal, modified look-up table for FreeSurfer.")

        LUTLABEL = ["tail", "head", "presubiculum", "subiculum", "ca1", "ca2", "ca3", "ca4", "dg", "ml"]

        LUTINDEX = [226, [233, 235, 237, 239, 245], 234, 236, 238, 240, 240, 242, 244, 246]

        LUTDICT = dict(zip(LUTLABEL, LUTINDEX))

        HSFLIST = [234, 236, 238, 240, 246]

    elif lut == "freesurfer-no_ml":
        LOGGER.info("Found internal, modified look-up table for FreeSurfer.")

        LUTLABEL = ["tail", "head", "presubiculum", "subiculum", "ca1", "ca2", "ca3", "ca4", "dg"]

        LUTINDEX = [226, [233, 235, 237, 239, 245], 234, 236, 238, 240, 240, 242, 244]

        LUTDICT = dict(zip(LUTLABEL, LUTINDEX))

        HSFLIST = [234, 236, 238, 240]

    elif lut == "ashs-penn_abc_3t_t2":
        LOGGER.info("Found internal, modified look-up table the Penn ABC-3T ASHS Atlas for T2-weighted MRI.")

        LUTLABEL = [
            "ca1",
            "ca2",
            "ca3",
            "ca4",
            "dg",
            "tail_orig",
            "subiculum",
            "presubiculum",
            "entorhinal",
            "ba35",
            "ba36",
            "parahippocampal",
            "head",
            "tail",
        ]

        LUTINDEX = [1, 2, 4, 3, 3, 5, 8, 8, 9, 10, 11, 12, 20, 5]

        LUTDICT = dict(zip(LUTLABEL, LUTINDEX))

        HSFLIST = [8, 1, 2, 4]

    elif lut == "ashs-umcutrecht_7t":
        LOGGER.info("Found internal, modified look-up table for ASHS UMC Utrecht 7T atlas.")

        LUTLABEL = [
            "entorhinal",
            "subiculum",
            "presubiculum",
            "ca1",
            "ca2",
            "dg",
            "ca3",
            "ca4",
            "cyst",
            "tail",
            "head",
        ]

        LUTINDEX = [1, 2, 2, 3, 4, 5, 6, 5, 7, 8, 20]

        LUTDICT = dict(zip(LUTLABEL, LUTINDEX))

        HSFLIST = [2, 3, 4, 6]

    elif os.path.isfile(lut):
        LOGGER.info("Found look-up table " + lut)

        lut = pandas.read_csv(
            lut,
            sep=" ",
            comment="#",
            header=None,
            skipinitialspace=True,
            skip_blank_lines=True,
            error_bad_lines=False,
            warn_bad_lines=True,
        )

        LUTDICT = dict(zip(lut[0], lut[1]))

        HSFLIST = list(lut[1])

    else:
        LUTDICT = dict()

        HSFLIST = []

    # add entries for tetra-labels

    LUTDICT["jointtail"] = 226
    LUTDICT["jointhead"] = 232
    LUTDICT["bndtail"] = 2260
    LUTDICT["bndhead"] = 2320
    LUTDICT["bndca4"] = 2420

    # return

    return LUTDICT, HSFLIST



"""
This module provides a function to create QC plots

"""

import logging
import os

import nibabel as nb
import numpy as np
import plotly.graph_objects as go
from lapy import TriaMesh, io
from lapy import plot as lpp
from plotly.subplots import make_subplots

from .get_levelsets import levelsetsTria

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS


def _sortLevelSets(LVL, dims, tol=1e-16):
    # create array of line segments
    tmpx = list()
    tmpy = list()

    for i in range(len(LVL[1][0])):
        tmpx.append(
            (
                LVL[0][0][LVL[1][0][i][0] - 1][dims[0]],
                LVL[0][0][LVL[1][0][i][1] - 1][dims[0]],
            )
        )
        tmpy.append(
            (
                LVL[0][0][LVL[1][0][i][0] - 1][dims[1]],
                LVL[0][0][LVL[1][0][i][1] - 1][dims[1]],
            )
        )

    tmpx = np.array(tmpx)
    tmpy = np.array(tmpy)

    # remove duplicate points
    tmpxy = np.unique(np.concatenate((tmpx, tmpy), axis=1), axis=0)
    tmpx = tmpxy[:, 0:2]
    tmpy = tmpxy[:, 2:4]

    # remove segments which are de-facto points
    tmpIdx = np.logical_or(np.abs(tmpx[:, 0] - tmpx[:, 1]) > tol, np.abs(tmpy[:, 0] - tmpy[:, 1]) > tol)
    tmpx = tmpx[tmpIdx, :]
    tmpy = tmpy[tmpIdx, :]

    # need to order array of line segments; whenever we encounter a
    # closed loop, we will already plot; otherwise, plot in the end
    sortIdx = np.array(range(0, len(tmpx)))

    tmpxSort = np.array(tmpx[sortIdx[0],], ndmin=2)
    tmpySort = np.array(tmpy[sortIdx[0],], ndmin=2)

    sortIdx = np.delete(sortIdx, sortIdx[0])

    while len(sortIdx) > 1:
        findIdx = np.array(
            np.where(
                np.logical_and(
                    np.abs(tmpx[sortIdx,] - tmpxSort[tmpxSort.shape[0] - 1, 1]) < tol,
                    np.abs(tmpy[sortIdx,] - tmpySort[tmpySort.shape[0] - 1, 1]) < tol,
                )
            ),
            ndmin=2,
        ).T

        # delete existing finds
        findIdxKeep = list()
        for k in range(findIdx.shape[0]):
            if not np.any(
                np.all(
                    np.logical_or(
                        tmpx[sortIdx[findIdx[k, 0]], 0] == tmpxSort,
                        tmpx[sortIdx[findIdx[k, 0]], 1] == tmpxSort,
                    ),
                    axis=1,
                )
            ):
                findIdxKeep.append(k)
        findIdx = findIdx[findIdxKeep,]

        if findIdx.shape[0] == 0:
            # reset (start new loop)
            tmpxSort = np.array(tmpx[sortIdx[0],], ndmin=2)
            tmpySort = np.array(tmpy[sortIdx[0],], ndmin=2)
            sortIdx = np.delete(sortIdx, 0)
        elif findIdx.shape[0] == 1:
            # add to current set
            if findIdx[0, 1] == 0:
                tmpxSort = np.append(
                    tmpxSort,
                    np.array(tmpx[sortIdx[findIdx[0, 0]], ::1], ndmin=2),
                    axis=0,
                )
                tmpySort = np.append(
                    tmpySort,
                    np.array(tmpy[sortIdx[findIdx[0, 0]], ::1], ndmin=2),
                    axis=0,
                )
            elif findIdx[0, 1] == 1:
                tmpxSort = np.append(
                    tmpxSort,
                    np.array(tmpx[sortIdx[findIdx[0, 0]], ::-1], ndmin=2),
                    axis=0,
                )
                tmpySort = np.append(
                    tmpySort,
                    np.array(tmpy[sortIdx[findIdx[0, 0]], ::-1], ndmin=2),
                    axis=0,
                )
            sortIdx = np.delete(sortIdx, findIdx[0, 0])
        elif findIdx.shape[0] > 1:
            # warning
            LOGGER.warning("A problem occurred with the surface overlays")
            break

    return tmpxSort, tmpySort


def qcPlots(params, stage=None):
    # get axes
    img = nb.load(params.FILENAME)
    ras2ras_tkr = img.header.get_vox2ras_tkr() @ img.header.get_ras2vox()
    ornts = nb.orientations.io_orientation(np.linalg.inv(ras2ras_tkr))
    scale_factor = 2
    tilt_factor = 1
    if ornts[0, 0] == 2:
        up = dict(x=ornts[0, 1], y=0, z=0)
    elif ornts[1, 0] == 2:
        up = dict(x=0, y=ornts[0, 1], z=0)
    elif ornts[2, 0] == 2:
        up = dict(x=0, y=0, z=ornts[0, 1])
    if ornts[0, 0] == 0:
        if ornts[1, 0] == 1:
            # SCA
            if params.HEMI == "lh":
                eye = dict(x=scale_factor * ornts[0, 1], y=tilt_factor * ornts[0, 1], z=tilt_factor * ornts[0, 1])
            else:
                eye = dict(x=-scale_factor * ornts[0, 1], y=-tilt_factor * ornts[0, 1], z=tilt_factor * ornts[0, 1])
        else:
            # SAC
            if params.HEMI == "lh":
                eye = dict(x=scale_factor * ornts[0, 1], y=tilt_factor * ornts[0, 1], z=tilt_factor * ornts[0, 1])
            else:
                eye = dict(x=-scale_factor * ornts[0, 1], y=tilt_factor * ornts[0, 1], z=-tilt_factor * ornts[0, 1])
    if ornts[1, 0] == 0:
        if ornts[0, 0] == 1:
            # CSA
            if params.HEMI == "lh":
                eye = dict(x=tilt_factor * ornts[0, 1], y=scale_factor * ornts[0, 1], z=tilt_factor * ornts[0, 1])
            else:
                eye = dict(x=-tilt_factor * ornts[0, 1], y=-scale_factor * ornts[0, 1], z=tilt_factor * ornts[0, 1])
        else:
            # ASC
            if params.HEMI == "lh":
                eye = dict(x=tilt_factor * ornts[0, 1], y=scale_factor * ornts[0, 1], z=tilt_factor * ornts[0, 1])
            else:
                eye = dict(x=tilt_factor * ornts[0, 1], y=-scale_factor * ornts[0, 1], z=-tilt_factor * ornts[0, 1])
    if ornts[2, 0] == 0:
        if ornts[1, 0] == 1:
            # ACS
            if params.HEMI == "lh":
                eye = dict(x=tilt_factor * ornts[0, 1], y=tilt_factor * ornts[0, 1], z=scale_factor * ornts[0, 1])
            else:
                eye = dict(x=tilt_factor * ornts[0, 1], y=-tilt_factor * ornts[0, 1], z=-scale_factor * ornts[0, 1])
        else:
            # CAS
            if params.HEMI == "lh":
                eye = dict(x=tilt_factor * ornts[0, 1], y=tilt_factor * ornts[0, 1], z=scale_factor * ornts[0, 1])
            else:
                eye = dict(x=-tilt_factor * ornts[0, 1], y=tilt_factor * ornts[0, 1], z=-scale_factor * ornts[0, 1])

    # mesh
    if params.internal.no_qc is False and stage == "mesh":
        triaMesh = TriaMesh.read_vtk(os.path.join(params.OUTDIR, params.HEMI + ".surf.vtk"))

        camera = dict(
            up=up,
            center=dict(x=0, y=0, z=0),
            eye=eye,
        )
        lpp.plot_tria_mesh(
            triaMesh,
            tcolor=[25, 25, 25],
            background_color="white",
            camera=camera,
            export_png=os.path.join(params.OUTDIR, "qc", params.HEMI + ".mesh.png"),
            no_display=True,
            scale_png=0.5,
        )

    # profile
    if params.internal.no_qc is False and stage == "profile":
        triaMesh = TriaMesh.read_vtk(os.path.join(params.OUTDIR, "tetra-cube", params.HEMI + ".rm.bnd.seam.rm.cut.vtk"))

        triaFunc = np.array(
            io.read_vfunc(
                os.path.join(
                    params.OUTDIR,
                    "tetra-cube",
                    params.HEMI + ".poisson1.rm.bnd.seam.rm.cut.psol",
                )
            )
        )

        #

        lvl2_0, lvl2i_0, lvl2j_0 = levelsetsTria(triaMesh.v, triaMesh.t, triaFunc, 0.10)
        lvl2_1, lvl2i_1, lvl2j_1 = levelsetsTria(triaMesh.v, triaMesh.t, triaFunc, 0.25)
        lvl2_2, lvl2i_2, lvl2j_2 = levelsetsTria(triaMesh.v, triaMesh.t, triaFunc, 0.50)
        lvl2_3, lvl2i_3, lvl2j_3 = levelsetsTria(triaMesh.v, triaMesh.t, triaFunc, 0.75)
        lvl2_4, lvl2i_4, lvl2j_4 = levelsetsTria(triaMesh.v, triaMesh.t, triaFunc, 0.90)

        #

        tmpxSort0, tmpySort0 = _sortLevelSets([lvl2_0, lvl2i_0, lvl2j_0], dims=[0, 2])
        tmpxSort1, tmpySort1 = _sortLevelSets([lvl2_1, lvl2i_1, lvl2j_1], dims=[0, 2])
        tmpxSort2, tmpySort2 = _sortLevelSets([lvl2_2, lvl2i_2, lvl2j_2], dims=[0, 2])
        tmpxSort3, tmpySort3 = _sortLevelSets([lvl2_3, lvl2i_3, lvl2j_3], dims=[0, 2])
        tmpxSort4, tmpySort4 = _sortLevelSets([lvl2_4, lvl2i_4, lvl2j_4], dims=[0, 2])

        #

        fig = make_subplots(rows=2, cols=3)

        fig.add_trace(go.Scatter(x=tmpxSort0[:, 0], y=tmpySort0[:, 0], mode="lines"), row=1, col=1)
        fig.add_trace(go.Scatter(x=tmpxSort1[:, 0], y=tmpySort1[:, 0], mode="lines"), row=1, col=2)
        fig.add_trace(go.Scatter(x=tmpxSort2[:, 0], y=tmpySort2[:, 0], mode="lines"), row=1, col=3)
        fig.add_trace(go.Scatter(x=tmpxSort3[:, 0], y=tmpySort3[:, 0], mode="lines"), row=2, col=1)
        fig.add_trace(go.Scatter(x=tmpxSort4[:, 0], y=tmpySort4[:, 0], mode="lines"), row=2, col=2)

        xmin = np.min(
            (
                np.min(tmpxSort0[:, 0]),
                np.min(tmpxSort1[:, 0]),
                np.min(tmpxSort2[:, 0]),
                np.min(tmpxSort3[:, 0]),
                np.min(tmpxSort4[:, 0]),
            )
        )
        xmax = np.max(
            (
                np.max(tmpxSort0[:, 0]),
                np.max(tmpxSort1[:, 0]),
                np.max(tmpxSort2[:, 0]),
                np.max(tmpxSort3[:, 0]),
                np.max(tmpxSort4[:, 0]),
            )
        )

        ymin = np.min(
            (
                np.min(tmpySort0[:, 0]),
                np.min(tmpySort1[:, 0]),
                np.min(tmpySort2[:, 0]),
                np.min(tmpySort3[:, 0]),
                np.min(tmpySort4[:, 0]),
            )
        )
        ymax = np.max(
            (
                np.max(tmpySort0[:, 0]),
                np.max(tmpySort1[:, 0]),
                np.max(tmpySort2[:, 0]),
                np.max(tmpySort3[:, 0]),
                np.max(tmpySort4[:, 0]),
            )
        )

        fig.update_layout(
            xaxis=dict(range=[xmin, xmax]),
            xaxis2=dict(range=[xmin, xmax]),
            xaxis3=dict(range=[xmin, xmax]),
            xaxis4=dict(range=[xmin, xmax]),
            xaxis5=dict(range=[xmin, xmax]),
            yaxis=dict(scaleanchor="x", range=[ymin, ymax]),
            yaxis2=dict(scaleanchor="x", range=[ymin, ymax]),
            yaxis3=dict(scaleanchor="x", range=[ymin, ymax]),
            yaxis4=dict(scaleanchor="x", range=[ymin, ymax]),
            yaxis5=dict(scaleanchor="x", range=[ymin, ymax]),
        )

        fig.update_layout(showlegend=False)

        fig.write_image(os.path.join(params.OUTDIR, "qc", params.HEMI + ".profile.png"))

    # hull
    if stage == "hull":
        triaMesh = TriaMesh.read_vtk(os.path.join(params.OUTDIR, "thickness", params.HEMI + ".hull.vtk"))

        camera = dict(
            up=up,
            center=dict(x=0, y=0, z=0),
            eye=eye,
        )
        lpp.plot_tria_mesh(
            triaMesh,
            tcolor=[25, 25, 25],
            background_color="white",
            camera=camera,
            export_png=os.path.join(params.OUTDIR, "qc", params.HEMI + ".hull.png"),
            no_display=True,
            scale_png=0.5,
        )

    # return
    return params



"""
This module provides a function to cut tetrahedral meshes

"""

import itertools as it
import os

import numpy as np
import scipy.linalg as la
import scipy.sparse as sp
from lapy import Solver, TetMesh, TriaMesh, io

from .utils.get_levelsets import levelsetsTetra

# ------------------------------------------------------------------------------
# AUXILIARY FUNCTIONS
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# _importData()


def _importData(tetraFile, tetraIdxFile):
    # read tetra mesh

    tetMesh = TetMesh.read_vtk(tetraFile)
    v4 = tetMesh.v
    t4 = tetMesh.t

    # read tetra index file

    l4 = np.array(io.read_vfunc(tetraIdxFile))

    # return

    return v4, t4, l4


# ------------------------------------------------------------------------------
# _exportData()


def _exportData(tetraCutOutFile, tetraCutOutFileFunc, tetraCutOutDir, hemi, v4c, t4c, i4c):
    # write out cut mesh and overlay

    tetMesh = TetMesh(v=v4c, t=t4c)

    TetMesh.write_vtk(tetMesh, tetraCutOutFile)

    io.write_vfunc(tetraCutOutFileFunc, i4c)

    # create boundary mesh and output

    triaMesh4cBnd = tetMesh.boundary_tria()

    triaMesh4cBnd.orient_()

    TriaMesh.write_vtk(triaMesh4cBnd, os.path.join(tetraCutOutDir, hemi + ".bnd.cut.vtk"))

    # remove all trias that share a vertex at the cut planes, because we want
    # to have an open boundary mesh; also remove free vertices

    t4cBndOpen = triaMesh4cBnd.t[np.sum(i4c[triaMesh4cBnd.t] == 1, axis=1) >= 1, :]

    triaMesh4cBndOpen = TriaMesh(v=v4c, t=t4cBndOpen)

    triaMesh4cBndOpen.orient_()

    TriaMesh.write_vtk(triaMesh4cBndOpen, os.path.join(tetraCutOutDir, hemi + ".open.bnd.cut.vtk"))

    # remove free vertices from boundary mesh and output

    triaMesh4cBnd.rm_free_vertices_()

    TriaMesh.write_vtk(triaMesh4cBnd, os.path.join(tetraCutOutDir, hemi + ".rm.bnd.cut.vtk"))

    # remove free vertices from open mesh and output ...

    triaMesh4cBndOpen.rm_free_vertices_()

    TriaMesh.write_vtk(triaMesh4cBndOpen, os.path.join(tetraCutOutDir, hemi + ".rm.open.bnd.cut.vtk"))

    # write out mapping between open bnd and v4c

    with open(os.path.join(tetraCutOutDir, hemi + ".rm.open.bnd.cut.lst"), "w") as f:
        [print(str(x), file=f) for x in np.unique(t4cBndOpen)]


# ------------------------------------------------------------------------------
# _preprocessData()


def _preprocessData(v4, t4, l4):
    tetMesh = TetMesh(v=v4, t=t4)

    tetMeshSolver = Solver(tetMesh)

    A = tetMeshSolver.stiffness

    M = tetMeshSolver.mass

    P = tetMeshSolver.poisson(dtup=(np.where(l4 != 0)[0], l4[np.where(l4 != 0)[0]]))

    return A, M, P


# ------------------------------------------------------------------------------
# _getTetra()


def _getTetra(aa, ii):
    # we compute which subset of three of neighbors of x has an adj mtx of
    # ones exclusively

    tt = list()

    for iii in range(len(ii)):
        n = np.where(aa[:, ii[iii]])[0]

        n = n[np.where(n != ii[iii])[0]]

        if len(n) >= 3:
            nt = list(it.combinations(n, 3))

            for ij in range(len(nt)):
                if aa[np.ix_(nt[ij], nt[ij])].all():
                    tt.append(np.sort(np.hstack((ii[iii], nt[ij]))))

    if len(tt) > 0:
        tt = np.unique(tt, axis=0)

    return tt


# ------------------------------------------------------------------------------
# _tetra26()


def _tetra26(tmp1, tmp2, tmp12, ind1, ind2, ind3, ind4, casevar):
    # create a local adj matrix: we know that tmp1, tmp2(1,:) and
    # tmp2(2,:) are connected within themselves
    ind = np.unique(
        np.vstack(
            (
                list(it.combinations(np.where(np.isin(tmp12, tmp1))[0], 2)),
                list(it.combinations(np.where(np.isin(tmp12, tmp2[0, :]))[0], 2)),
                list(it.combinations(np.where(np.isin(tmp12, tmp2[1, :]))[0], 2)),
            )
        ),
        axis=0,
    )
    a = np.zeros((tmp12.shape[0], tmp12.shape[0]))
    a[ind[:, 0], ind[:, 1]] = 1
    a = a + a.transpose()
    a = a + np.diag(np.ones(a.shape[0]))

    # add what we know
    if casevar == 1:
        # 1-3, 1-4, 2-5, 2-6
        a[0, 2] = 1
        a[2, 0] = 1
        a[0, 3] = 1
        a[3, 0] = 1
        a[1, 4] = 1
        a[4, 1] = 1
        a[1, 5] = 1
        a[5, 1] = 1

    elif casevar == 2:
        # 1-3, 1-5, 2-4, 2-6
        a[0, 2] = 1
        a[2, 0] = 1
        a[0, 4] = 1
        a[4, 0] = 1
        a[1, 3] = 1
        a[3, 1] = 1
        a[1, 5] = 1
        a[5, 1] = 1

    # add indices supplied by arguments
    a[ind1, ind2] = 1
    a[ind2, ind1] = 1
    a[ind3, ind4] = 1
    a[ind4, ind3] = 1

    # rewiring of the cut plane necessary if 1-6, 2-3
    if ind1 == 1 and ind2 == 2 and ind3 == 0 and ind4 == 5:
        a[3, 4] = 0
        a[4, 3] = 0
        a[2, 5] = 1
        a[5, 2] = 1

    # get tetra
    tetra = tmp12[_getTetra(a, range(0, 6))]

    # return
    return tetra, a


# ------------------------------------------------------------------------------
# _tetra33()


def _tetra33(tmp1, tmp2, tmp12, ind1, ind2, ind3, v, A):
    # create a local adj matrix: we know that tmp1 and tmp2 are
    # connected within themselves, and we know from the levelsetsTetra
    # code that tmp1(1) is connected with tmp2(1), tmp1(2) is connected
    # with tmp2(2), and tmp1(3) is connected with tmp2(3)
    a = (
        la.block_diag(np.ones((tmp1.shape[0], tmp1.shape[0])), np.ones((tmp1.shape[0], tmp1.shape[0])))
        + np.diag(np.ones(tmp1.shape[0]), k=3)
        + np.diag(np.ones(tmp1.shape[0]), k=-3)
    )
    a = a + np.diag(np.ones(a.shape[0]))

    # now add those edges as specified by the arguments
    a[ind1[0], ind1[1]] = 1
    a[ind1[1], ind1[0]] = 1
    a[ind2[0], ind2[1]] = 1
    a[ind2[1], ind2[0]] = 1
    a[ind3[0], ind3[1]] = 1
    a[ind3[1], ind3[0]] = 1

    # get all tetras
    tetra = tmp12[_getTetra(a, range(0, 6))]

    # need to add a point in two special cases:
    if (ind1[0] == 0 and ind1[1] == 5 and ind2[0] == 2 and ind2[1] == 4 and ind3[0] == 1 and ind3[1] == 3) or (
        ind1[0] == 2 and ind1[1] == 3 and ind2[0] == 1 and ind2[1] == 5 and ind3[0] == 0 and ind3[1] == 4
    ):
        # compute midpoints on each edge
        p1 = v[tmp1[0], :] + np.divide([v[tmp2[0], :] - v[tmp1[0], :]], 2)
        p2 = v[tmp1[1], :] + np.divide([v[tmp2[1], :] - v[tmp1[1], :]], 2)
        p3 = v[tmp1[2], :] + np.divide([v[tmp2[2], :] - v[tmp1[2], :]], 2)

        # compute center of triangle (p1, p2, p3)
        pc = np.divide((p1 + p2 + p3), 3)

        # add point to the list of vertices
        v = np.concatenate((v, pc), axis=0)

        # create new list of points
        tmp123 = np.hstack((tmp1, tmp2, np.shape(v)[0] - 1))

        # append a
        a = np.vstack((a, np.ones((1, np.shape(a)[1]))))
        a = np.hstack((a, np.ones((np.shape(a)[0], 1))))

        # get tetra
        tetra = tmp123[_getTetra(a, range(0, 7))]

        # add one row/col to A and set diag elem to 1
        A = sp.coo_matrix(
            (
                np.ones(np.shape(A.nonzero())[1] + 1),
                (np.hstack((A.nonzero()[0], np.shape(v)[0] - 1)), np.hstack((A.nonzero()[1], np.shape(v)[0] - 1))),
            ),
            shape=(np.shape(v)[0], np.shape(v)[0]),
        )

    # return
    return tetra, a, v, A


# ------------------------------------------------------------------------------
# MAIN FUNCTION
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# cutTetra()


def cutTetra(params):
    """
    This is a function to cut tetrahedral meshes

    """

    # -------------------------------------------------------------------------
    # message
    # -------------------------------------------------------------------------

    print()
    print("--------------------------------------------------------------------------------")
    print("Cutting tetrahedral mesh")
    print()

    # -------------------------------------------------------------------------
    # evaluate input
    # -------------------------------------------------------------------------

    tetraFile = os.path.join(params.OUTDIR, params.HEMI + ".tetra.vtk")

    tetraIdxFile = os.path.join(params.OUTDIR, "tetra-cut", params.HEMI + ".tetra-remove_bnd.psol")

    tetraCutOutFile = os.path.join(params.OUTDIR, params.HEMI + ".cut.vtk")

    tetraCutOutFileFunc = os.path.join(params.OUTDIR, "tetra-cut", params.HEMI + ".cut.psol")

    tetraCutOutDir = os.path.join(params.OUTDIR, "tetra-cut")

    # -------------------------------------------------------------------------
    # get data
    # -------------------------------------------------------------------------

    v4, t4, l4idx = _importData(tetraFile, tetraIdxFile)

    l4idx[np.where(l4idx < params.LUTDICT["bndtail"])[0]] = 0
    l4idx[np.where(l4idx == params.LUTDICT["bndca4"])[0]] = 0
    l4idx[np.where(l4idx == params.LUTDICT["bndhead"])[0]] = -1
    l4idx[np.where(l4idx == params.LUTDICT["bndtail"])[0]] = 1

    # -------------------------------------------------------------------------
    # preprocess data
    # -------------------------------------------------------------------------

    Ap, Mp, l4 = _preprocessData(v4, t4, l4idx)

    # -------------------------------------------------------------------------
    # create a smooth boundary
    # -------------------------------------------------------------------------

    # compute level sets
    vlTail, llTail, ilTail, jlTail, olTail = levelsetsTetra(v4, t4, l4, params.internal.cut_range[0])
    vlHead, llHead, ilHead, jlHead, olHead = levelsetsTetra(v4, t4, l4, params.internal.cut_range[1])

    # find all tetras that do not exceed the cutting criteria
    t4c1 = t4[
        np.where(
            np.sum(
                np.isin(t4, np.where((l4 > params.internal.cut_range[0]) & (l4 < params.internal.cut_range[1]))), axis=1
            )
            == 4
        ),
        :,
    ][0]

    # add new points to v and generate new triangles
    v4c = np.concatenate((v4, vlTail[0], vlHead[0]), axis=0)
    t4c = t4c1

    # create adjacency matrix
    idx = np.concatenate(
        (t4c[:, [0, 1]], t4c[:, [0, 2]], t4c[:, [0, 3]], t4c[:, [1, 2]], t4c[:, [1, 3]], t4c[:, [2, 3]]), axis=0
    )
    A = sp.eye(np.shape(v4c)[0])
    A = A + sp.coo_matrix(
        (np.ones((np.shape(idx)[0])), (idx[:, 0], idx[:, 1])), shape=(np.shape(v4c)[0], np.shape(v4c)[0])
    )
    A = A + A.transpose()
    A[A > 0] = 1
    A.tolil()

    # -------------------------------------------------------------------------
    # iterate through all tetras on the level set labelTail
    # -------------------------------------------------------------------------

    for i in range(len(ilTail[0])):
        # check which points of the original tetra are not exceeding the
        # plane, can be 1 or 2 or 3 points
        tmp1 = t4[ilTail[0][i], olTail[0][i][:]]

        # find points on the level set, can be 1x3 or 2x3 points, where 1x3
        # only occurs with tmp1==1 and tmp1==3, and 2x3 only with tmp1==2. If
        # there are 2x3 points, only 4 of them are unique.
        # Note that we need to change from 1-based to 0-based indices here,
        # therefore we add '-1' at the end.
        tmp2 = (llTail[0][np.where(jlTail[0] == ilTail[0][i])[0][:], :] + np.shape(v4)[0] - 1).squeeze()

        # the general strategy is to test out all degrees of freedom that we
        # have to see if there is something fitting. That are 1 deg for 1|3,
        # 2 deg for 2|6, and 3*2 deg for 3|3. this does not include any
        # rewiring or adding points.

        if (np.size(tmp1) == 1) & (np.size(tmp2) == 3):
            # no check against A should be necessary here, because we have no
            # degrees of freedom in constructing the tetra. in other words,
            # this tetra should always fit.

            # concatenate tmp1 and tmp2. this is already the tetra.
            tetra = np.hstack((tmp1, tmp2))[np.newaxis, :]

        elif (np.size(tmp1) == 2) & (np.size(tmp2) == 6):
            # concatenate tmp1 and tmp2. drop duplicate points.
            idx = np.unique(tmp2.flatten(), return_index=True)[1]
            tmp12 = np.hstack((tmp1, [tmp2.flatten()[idxi] for idxi in sorted(idx)]))

            # compute tetras and local adjacency matrix for various edge
            # insertion options
            tetra21, A21 = _tetra26(tmp1, tmp2, tmp12, 0, 3, 0, 5, 2)
            tetra22, A22 = _tetra26(tmp1, tmp2, tmp12, 0, 3, 1, 4, 2)
            tetra23, A23 = _tetra26(tmp1, tmp2, tmp12, 1, 2, 0, 5, 2)
            tetra24, A24 = _tetra26(tmp1, tmp2, tmp12, 1, 2, 1, 4, 2)

            # check which variant is compatible with A: all edges from the
            # relevant subset of A must be present in the local adj mtx.
            # However, the updated local adj mtx may contain new edges that
            # will be added to A.
            if not ((A21 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra21
            elif not ((A22 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra22
            elif not ((A23 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra23
            elif not ((A24 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra24

        elif (np.size(tmp1) == 3) & (np.size(tmp2) == 3):
            # concatenate tmp1 and tmp2
            tmp12 = np.hstack((tmp1, tmp2))

            # compute tetras and local adjacency matrix for various edge
            # insertion options
            tetra111, a111, v4c111, A111 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [2, 4], [0, 4], v4c, A)
            tetra112, a112, v4c112, A112 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [2, 4], [1, 3], v4c, A)
            tetra121, a121, v4c121, A121 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [1, 5], [0, 4], v4c, A)
            tetra122, a122, v4c122, A122 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [1, 5], [1, 3], v4c, A)
            tetra211, a211, v4c211, A211 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [2, 4], [0, 4], v4c, A)
            tetra212, a212, v4c212, A212 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [2, 4], [1, 3], v4c, A)
            tetra221, a221, v4c221, A221 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [1, 5], [0, 4], v4c, A)
            tetra222, a222, v4c222, A222 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [1, 5], [1, 3], v4c, A)

            # check which variant is compatible with A: all edges from the
            # relevant subset of A must be present in the local adj mtx.
            # However, the updated local adj mtx may contain new edges that
            # will be added to A.
            if not ((a111[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra111
                v4c = v4c111
                A = A111.tolil()
            elif not ((a112[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra112
                v4c = v4c112
                A = A112.tolil()
            elif not ((a121[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra121
                v4c = v4c121
                A = A121.tolil()
            elif not ((a122[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra122
                v4c = v4c122
                A = A122.tolil()
            elif not ((a211[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra211
                v4c = v4c211
                A = A211.tolil()
            elif not ((a212[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra212
                v4c = v4c212
                A = A212.tolil()
            elif not ((a221[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra221
                v4c = v4c221
                A = A221.tolil()
            elif not ((a222[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra222
                v4c = v4c222
                A = A222.tolil()

        # update adjacency matrix

        for k in range(np.shape(tetra)[0]):
            A[tetra[k, 0], tetra[k, 1]] = 1
            A[tetra[k, 1], tetra[k, 0]] = 1

            A[tetra[k, 0], tetra[k, 2]] = 1
            A[tetra[k, 2], tetra[k, 0]] = 1

            A[tetra[k, 0], tetra[k, 3]] = 1
            A[tetra[k, 3], tetra[k, 0]] = 1

            A[tetra[k, 1], tetra[k, 2]] = 1
            A[tetra[k, 2], tetra[k, 1]] = 1

            A[tetra[k, 1], tetra[k, 3]] = 1
            A[tetra[k, 3], tetra[k, 1]] = 1

            A[tetra[k, 2], tetra[k, 3]] = 1
            A[tetra[k, 3], tetra[k, 2]] = 1

        # update list of tetras

        t4c = np.concatenate((t4c, tetra), axis=0)

    # -------------------------------------------------------------------------
    # iterate through all tetras on the level set labelHead
    # -------------------------------------------------------------------------

    for i in range(len(ilHead[0])):
        # check which points of the original tetra are not exceeding the
        # plane, can be 1 or 2 or 3 points; note the setdiff --> outlying
        # points are defined from the other direction as compared to vTail.
        tmp1 = t4[ilHead[0][i], np.setdiff1d((0, 1, 2, 3), olHead[0][i][:])]

        # find points on the level set, can be 1x3 or 2x3 points, where 1x3
        # only occurs with tmp1==1 and tmp1==3, and 2x3 only with tmp1==2. If
        # there are 2x3 points, only 4 of them are unique.
        # Note that we need to change from 1-based to 0-based indices here,
        # therefore we add '-1' at the end.
        tmp2 = (
            llHead[0][np.where(jlHead[0] == ilHead[0][i])[0][:], :] + np.shape(v4)[0] + np.shape(vlTail[0])[0] - 1
        ).squeeze()

        # the general strategy is to test out all degrees of freedom that we
        # have to see if there is something fitting. That are 1 deg for 1|3,
        # 2 deg for 2|6, and 3*2 deg for 3|3. this does not include any
        # rewiring or adding points.

        if (np.size(tmp1) == 1) & (np.size(tmp2) == 3):
            # no check against A should be necessary here, because we have no
            # degrees of freedom in constructing the tetra. in other words,
            # this tetra should always fit.

            # concatenate tmp1 and tmp2. this is already the tetra.
            tetra = np.hstack((tmp1, tmp2))[np.newaxis, :]

        elif (np.size(tmp1) == 2) & (np.size(tmp2) == 6):
            # concatenate tmp1 and tmp2. drop duplicate points.
            idx = np.unique(tmp2.flatten(), return_index=True)[1]
            tmp12 = np.hstack((tmp1, [tmp2.flatten()[idxi] for idxi in sorted(idx)]))

            # compute tetras and local adjacency matrix for various edge
            # insertion options
            tetra11, A11 = _tetra26(tmp1, tmp2, tmp12, 0, 4, 0, 5, 1)
            tetra12, A12 = _tetra26(tmp1, tmp2, tmp12, 0, 4, 1, 3, 1)
            tetra13, A13 = _tetra26(tmp1, tmp2, tmp12, 1, 2, 0, 5, 1)
            tetra14, A14 = _tetra26(tmp1, tmp2, tmp12, 1, 2, 1, 3, 1)

            # check which variant is compatible with A: all edges from the
            # relevant subset of A must be present in the local adj mtx.
            # However, the updated local adj mtx may contain new edges that
            # will be added to A.

            if not ((A11 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra11
            elif not ((A12 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra12
            elif not ((A13 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra13
            elif not ((A14 - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra14

        elif (np.size(tmp1) == 3) & (np.size(tmp2) == 3):
            # concatenate tmp1 and tmp2
            tmp12 = np.hstack((tmp1, tmp2))

            # compute tetras and local adjacency matrix for various edge
            # insertion options
            tetra111, a111, v4c111, A111 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [2, 4], [0, 4], v4c, A)
            tetra112, a112, v4c112, A112 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [2, 4], [1, 3], v4c, A)
            tetra121, a121, v4c121, A121 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [1, 5], [0, 4], v4c, A)
            tetra122, a122, v4c122, A122 = _tetra33(tmp1, tmp2, tmp12, [0, 5], [1, 5], [1, 3], v4c, A)
            tetra211, a211, v4c211, A211 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [2, 4], [0, 4], v4c, A)
            tetra212, a212, v4c212, A212 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [2, 4], [1, 3], v4c, A)
            tetra221, a221, v4c221, A221 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [1, 5], [0, 4], v4c, A)
            tetra222, a222, v4c222, A222 = _tetra33(tmp1, tmp2, tmp12, [2, 3], [1, 5], [1, 3], v4c, A)

            # check which variant is compatible with A: all edges from the
            # relevant subset of A must be present in the local adj mtx.
            # However, the updated local adj mtx may contain new edges that
            # will be added to A.

            if not ((a111[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra111
                v4c = v4c111
                A = A111.tolil()
            elif not ((a112[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra112
                v4c = v4c112
                A = A112.tolil()
            elif not ((a121[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra121
                v4c = v4c121
                A = A121.tolil()
            elif not ((a122[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra122
                v4c = v4c122
                A = A122.tolil()
            elif not ((a211[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra211
                v4c = v4c211
                A = A211.tolil()
            elif not ((a212[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra212
                v4c = v4c212
                A = A212.tolil()
            elif not ((a221[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra221
                v4c = v4c221
                A = A221.tolil()
            elif not ((a222[0:6, 0:6] - A[tmp12, :][:, tmp12].todense()) < 0).any():
                tetra = tetra222
                v4c = v4c222
                A = A222.tolil()

        # update adjacency matrix

        for k in range(np.shape(tetra)[0]):
            A[tetra[k, 0], tetra[k, 1]] = 1
            A[tetra[k, 1], tetra[k, 0]] = 1

            A[tetra[k, 0], tetra[k, 2]] = 1
            A[tetra[k, 2], tetra[k, 0]] = 1

            A[tetra[k, 0], tetra[k, 3]] = 1
            A[tetra[k, 3], tetra[k, 0]] = 1

            A[tetra[k, 1], tetra[k, 2]] = 1
            A[tetra[k, 2], tetra[k, 1]] = 1

            A[tetra[k, 1], tetra[k, 3]] = 1
            A[tetra[k, 3], tetra[k, 1]] = 1

            A[tetra[k, 2], tetra[k, 3]] = 1
            A[tetra[k, 3], tetra[k, 2]] = 1

        # update list of tetras

        t4c = np.concatenate((t4c, tetra), axis=0)

    #
    t4c = np.unique(t4c, axis=0)

    #
    i4c = np.concatenate(
        (
            1 * np.ones(np.shape(v4)[0]),  # original points (includes exterior and interior points)
            params.LUTDICT["jointtail"] * np.ones(np.shape(vlTail[0])[0]),  # labelTail points
            params.LUTDICT["jointhead"] * np.ones(np.shape(vlHead[0])[0]),  # labelHead points
            2
            * np.ones(
                np.shape(v4c)[0] - np.shape(v4)[0] - np.shape(vlTail[0])[0] - np.shape(vlHead[0])[0]
            ),  # new points
        )
    )

    # -------------------------------------------------------------------------
    # export data
    # -------------------------------------------------------------------------

    _exportData(tetraCutOutFile, tetraCutOutFileFunc, tetraCutOutDir, params.HEMI, v4c, t4c, i4c)

    # -------------------------------------------------------------------------
    # return
    # -------------------------------------------------------------------------

    return params



"""
This module provides a function to map values from a volume to the midsurface.

"""

import argparse
import logging
import os
import subprocess
import sys

import nibabel as nb
import numpy as np
import pandas as pd
from lapy import TriaMesh, io
from scipy import stats as st

# ==============================================================================
# LOGGING

LOGGER = logging.getLogger(__name__)

# ==============================================================================
# FUNCTIONS

# ------------------------------------------------------------------------------
# AUXILIARY FUNCTIONS
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# get_help()


def get_help(print_help=True, return_help=False):
    """
    a function to return a help message

    """

    # define helptext
    HELPTEXT = """
    SUMMARY

    This is a script for mapping volume-based data onto the hippocampal
    midsurface.

    USAGE

    The script requires four arguments:

    --volume <input volume>
    --surface <input surface>
    --label <input label file>
    --table <input csv table>
    --suffix <suffix>

    The following arguments are optional and determine which output files will
    be created:

    --integrate <none|mode|median|mean|min|max> (default: none)
    --select <0, 1, ..., n> (default: all)
    --interp <nearest|cubic> (default: nearest)
    --writePSOL
    --writeMGH

    EXAMPLES

    Example for mapping at the vertices of the midsurface:

    python3 mapValues.py --volume /my/path/to/<MY_VOLUME.mgz> \\
        --surface /my/path/to/<SUBJECT_ID>/thickness/lh.mid-surface.vtk \\
        --label /my/path/to/<MY_LABEL_FILE.mgz> \\
        --table /my/path/to/<SUBJECT_ID>/thickness/lh.mid-surface.csv \\
        --suffix MY_SUFFIX --writePSOL --writeMGH

    This will sample the vertices of the midsurface at the corresponding
    positions in the given volume. This mapping requires the '[lr]h.mid-
    surface.vtk' surface and the '[lr]h.mid-surface.csv' index table that are
    created by the hippocampal thickness tools. It also requires the original
    label file that contains the subfield labels. In some cases it will be
    necessary to reslice the input volume to match the input label file: in
    this case, there are two interpolation options, nearest neighbor and cubic.
    Nearest neighbor should be used for labels, and cubic for continuous
    volumetric data. The resliced file will be saved in the same directory as
    the input file. The suffix argument will be used to identify the output
    files and can be chosen arbitrarily. PSOL and MGH output files will be
    written to the same directory as the input surface.

    Example for mapping at the vertices of the streamlines:

    python3 mapValues.py --volume /my/path/to/<MY_VOLUME.mgz> \\
        --surface /my/path/to/<SUBJECT_ID>/thickness/lh.grid-lines-z.vtk \\
        --label /my/path/to/<SUBJECT_ID>/mri/lh.hippoAmygLabels-T1.v20.mgz \\
        --table /my/path/to/<SUBJECT_ID>/thickness/lh.grid-lines.csv \\
        --suffix MY_SUFFIX --integrate mean --select 0 1 2 3 4 5 \\
        --interp cubic --writePSOL --writeMGH

    This will sample the vertices of the interior/exterior streamlines at the
    corresponding positions in the given volume. That is, multiple vertices per
    streamlines will be integrated and projected onto the corresponding point
    of the midsurface. This mapping requires the '[lr]h.grid-lines-z.vtk' file
    and the '[lr]h.grid-lines.csv' index table that is created by the hippo-
    campal thickness tools. Also, the 'integrate' argument is required. In this
    example, the sampled values will be averaged ('mean') per streamline. The
    select argument can be used to restrict the integration to a subset of the
    integration points (0 ... 10 for the default z direction, 5 being the
    midsurface). We assume that our input volume contains continuous volumetric
    data, and hence use cubic interpolation. The suffix will be used to
    identify the output files and can be chosen arbitrarily. PSOL and MGH
    output files will be written to the same directory as the input surface.

    """

    if print_help:
        print(HELPTEXT)

    if return_help:
        return HELPTEXT


# ------------------------------------------------------------------------------
# parse_arguments


def _parse_arguments():
    """
    Command Line Options Parser:
    initiate the option parser and return the parsed object
    """

    # message
    print("\nReading input options ...")

    # setup parser
    parser = argparse.ArgumentParser(description="", add_help=False)

    # help text
    h_IN_VOL = "input volume"
    h_IN_SURF = "input surface"
    h_IN_LABEL = "input label file"
    h_IN_INDICES = "table with indices"
    h_IN_SUFFIX = "suffix for output files"
    h_writePSOL = "write out PSOL files"
    h_writeMGH = "write out MGH files"
    h_writeANNOT = "write out ANNOT files"
    h_integrate = "write out integrated values (default: none)"
    h_select = "select inegration points (default: all)"
    h_interp = "type of interpolation (cubic or nearest; default: nearest)"

    # required arguments
    required = parser.add_argument_group("required arguments")

    required.add_argument(
        "--volume",
        dest="IN_VOL",
        help=h_IN_VOL,
        default=None,
        required=False,
        metavar="<file>",
    )
    required.add_argument(
        "--surface",
        dest="IN_SURF",
        help=h_IN_SURF,
        default=None,
        required=False,
        metavar="<file>",
    )
    required.add_argument(
        "--label",
        dest="IN_LABEL",
        help=h_IN_LABEL,
        default=None,
        required=False,
        metavar="<file>",
    )
    required.add_argument(
        "--table",
        dest="IN_INDICES",
        help=h_IN_INDICES,
        default=None,
        required=False,
        metavar="<file>",
    )
    required.add_argument(
        "--suffix",
        dest="IN_SUFFIX",
        help=h_IN_SUFFIX,
        default=None,
        required=False,
        metavar="<string>",
    )

    # optional arguments
    optional = parser.add_argument_group("optional arguments")

    optional.add_argument(
        "--integrate",
        dest="INTEGRATE",
        help=h_integrate,
        default="none",
        required=False,
        metavar="<none|mode|median|mean|min|max>",
    )
    optional.add_argument(
        "--select",
        dest="SELECT",
        help=h_select,
        default=None,
        nargs="+",
        required=False,
        metavar="<list>",
    )
    optional.add_argument(
        "--interp",
        dest="INTERP",
        help=h_interp,
        default="nearest",
        required=False,
        metavar="<nearest|cubic>",
    )
    optional.add_argument(
        "--writePSOL",
        dest="writePSOL",
        help=h_writePSOL,
        default=False,
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--writeMGH",
        dest="writeMGH",
        help=h_writeMGH,
        default=False,
        action="store_true",
        required=False,
    )
    optional.add_argument(
        "--writeANNOT",
        dest="writeANNOT",
        help=argparse.SUPPRESS,
        default=False,
        action="store_true",
        required=False,
    )  # help=h_writeANNOT # currently hidden, might be added later

    # define help
    help = parser.add_argument_group("getting help")

    help.add_argument("--help", help="display this help message and exit", action="help")
    help.add_argument(
        "--more-help",
        dest="more_help",
        help="display extensive help message and exit",
        default=False,
        action="store_true",
        required=False,
    )

    # check if there are any inputs; if not, print help and exit
    if len(sys.argv) == 1:
        args = parser.parse_args(["--help"])
    else:
        args = parser.parse_args()

    return args


# ------------------------------------------------------------------------------
# check arguments


def _check_arguments(options):
    """
    an internal function to check input arguments

    """

    # more help

    if options.more_help is True:
        get_help()
        sys.exit(0)

    # check environment variables

    if os.environ.get("FREESURFER_HOME") is None:
        LOGGER.info("ERROR: need to set the FREESURFER_HOME environment variable and to source FREESURFER.")
        sys.exit(1)

    # required arguments

    if options.IN_VOL is None:
        LOGGER.info("The --volume option is required. See --help for details.")
        sys.exit(1)

    if options.IN_SURF is None:
        LOGGER.info("The --surface option is required. See --help for details.")
        sys.exit(1)

    if options.IN_LABEL is None:
        LOGGER.info("The --label option is required. See --help for details.")
        sys.exit(1)

    if options.IN_INDICES is None:
        LOGGER.info("The --table option is required. See --help for details.")
        sys.exit(1)

    if options.IN_SUFFIX is None:
        LOGGER.info("The --suffix option is required. See --help for details.")
        sys.exit(1)

    # change formats

    if options.SELECT is not None:
        options.SELECT = np.array(options.SELECT).astype(int)

    # add params

    options.params = None

    # return

    return options


# ------------------------------------------------------------------------------
# MAIN FUNCTION
# ------------------------------------------------------------------------------


def mapValues(
    params,
    IN_VOL=None,
    IN_SURF=None,
    IN_LABEL=None,
    IN_INDICES=None,
    IN_SUFFIX="hsf",
    INTEGRATE="mode",
    SELECT=None,
    INTERP="nearest",
    writePSOL=False,
    writeMGH=False,
    writeANNOT=False,
):
    # message

    print()
    print("--------------------------------------------------------------------------------")
    print("Mapping values")
    print()

    # -------------------------------------------------------
    # get params

    if params is not None:
        # these are the default settings for processing within the hippocampal
        # thickness toolbox; will override any other settings; set 'params=None'
        # for custom processing.
        IN_VOL = os.path.join(params.OUTDIR, params.HEMI + ".labels.mgz")
        IN_SURF = os.path.join(params.OUTDIR, "thickness", params.HEMI + ".mid-surface.vtk")  # or: grid-lines-z.vtk
        IN_LABEL = params.FILENAME
        IN_INDICES = os.path.join(params.OUTDIR, "thickness", params.HEMI + ".mid-surface.csv")  # or: grid-lines.csv
        IN_SUFFIX = "hsf"

        writePSOL = params.internal.mapValuesWritePSOL
        writeMGH = params.internal.mapValuesWriteMGH
        writeANNOT = params.internal.mapValuesWriteANNOT
        INTEGRATE = params.internal.mapValuesIntegrate
        SELECT = params.internal.mapValuesSelect
        INTERP = params.internal.mapValuesInterp

    # -------------------------------------------------------
    # check for differences

    cmd = (
        os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_diff")
        + " "
        + "--notallow-acq"
        + " "
        + "--notallow-prec"
        + " "
        + "--notallow-pix"
        + " "
        + IN_VOL
        + " "
        + IN_LABEL
    )
    # + " " + "--log" + " " + os.path.splitext(IN_VOL)[0] + ".diff"

    print(cmd)

    subproc = subprocess.run(cmd.split(), capture_output=True)

    print(subproc.returncode)

    # reslice input if necessary
    if subproc.returncode == 1:
        sys.exit(1)
    elif subproc.returncode > 100:
        if os.path.isdir(os.path.join(os.path.dirname(IN_VOL), "labels")):
            RESLICED_VOL = os.path.join(
                os.path.dirname(IN_VOL),
                "labels",
                os.path.splitext(os.path.basename(IN_VOL))[0] + "_reslicedToHighresHC" + os.path.splitext(IN_VOL)[1],
            )
        else:
            RESLICED_VOL = os.path.splitext(IN_VOL)[0] + "_reslicedToHighresHC" + os.path.splitext(IN_VOL)[1]

        cmd = (
            os.path.join(os.environ.get("FREESURFER_HOME"), "bin", "mri_convert")
            + " -rl "
            + IN_LABEL
            + " -rt "
            + INTERP
            + " "
            + IN_VOL
            + " "
            + RESLICED_VOL
        )

        print(cmd)

        subprocess.run(cmd.split(), capture_output=True)

    # read files
    if subproc.returncode > 100:
        vol = nb.freesurfer.load(RESLICED_VOL)
    else:
        vol = nb.freesurfer.load(IN_VOL)

    surf = TriaMesh.read_vtk(IN_SURF)

    # get data
    mat = vol.header.get_vox2ras_tkr()
    dat = vol.get_fdata()
    ind = np.array(np.nonzero(dat)).transpose()

    # variant 1: do it in surface RAS space (disadvantage: need to do distance
    # computation for many unnecessary values)
    # from scipy import spatial as sp
    # coordXYZ = np.concatenate((ind, np.ones((len(ind), 1))), axis=1)
    # coordSurfRAS = np.matmul(coordXYZ, mat.transpose())[:, 0:3]
    # dst = sp.distance_matrix(surf.v, coordSurfRAS)

    # variant 2: do it in XYZ space (easier lookup for relevant mid-surface
    # vertices)
    vtcs = np.concatenate((surf.v, np.ones((len(surf.v), 1))), axis=1)
    vtcsXYZ = np.matmul(vtcs, np.linalg.inv(mat).transpose())[:, 0:3]
    vtcsXYZ = np.round(vtcsXYZ).astype("int")
    lookup = dat[vtcsXYZ[:, 0], vtcsXYZ[:, 1], vtcsXYZ[:, 2]]

    # integrate along z-axis
    indices = np.array(pd.read_csv(IN_INDICES, header=None, index_col=None))
    indices3D = indices[:, 0:3].astype(int)
    lookup3D = np.full((np.max(indices3D, axis=0) + 1), np.nan)
    for i in range(0, len(lookup)):
        lookup3D[indices3D[i, 0], indices3D[i, 1], indices3D[i, 2]] = lookup[i]

    # if no selection is specified, set SELECT to all available indices
    if SELECT is None:
        SELECT = np.unique(indices3D[:, 2])

    if INTEGRATE == "mean":
        integr = np.nanmean(lookup3D[:, :, SELECT], axis=2)
    elif INTEGRATE == "median":
        integr = np.nanmedian(lookup3D[:, :, SELECT], axis=2)
    elif INTEGRATE == "mode":
        integr = st.mode(lookup3D[:, :, SELECT], axis=2)[0]
    elif INTEGRATE == "max":
        integr = np.nanmax(lookup3D[:, :, SELECT], axis=2)
    elif INTEGRATE == "min":
        integr = np.nanmin(lookup3D[:, :, SELECT], axis=2)
    elif INTEGRATE == "nonzeromin":
        lookup3D[lookup3D == 0] = np.Inf
        integr = np.nanmin(lookup3D[:, :, SELECT], axis=2)
    elif INTEGRATE == "none":
        if len(SELECT) > 1:
            LOGGER.info("Error: cannot use --integrate none with multiple sampling points, exiting.")
            sys.exit(1)
        else:
            integr = lookup3D[:, :, SELECT]

    integr = np.vstack(
        (
            np.indices(integr.shape)[0].flatten(),
            np.indices(integr.shape)[1].flatten(),
            integr.flatten(),
        )
    ).transpose()
    integr = pd.DataFrame(integr).sort_values([0, 1])

    # output
    OUT_CSV = IN_SURF.replace(".vtk", "." + IN_SUFFIX + ".csv")
    pd.DataFrame(integr).to_csv(OUT_CSV, header=False, index=False)

    if writePSOL is True:
        if (
            INTEGRATE == "mean"
            or INTEGRATE == "median"
            or INTEGRATE == "mode"
            or INTEGRATE == "max"
            or INTEGRATE == "min"
        ):
            OUT_PSOL_INTEGR = IN_SURF.replace(".vtk", "." + IN_SUFFIX + "-integrated.psol")
            io.write_vfunc(OUT_PSOL_INTEGR, np.asarray(integr)[:, 2])
        else:
            OUT_PSOL = IN_SURF.replace(".vtk", "." + IN_SUFFIX + ".psol")
            io.write_vfunc(OUT_PSOL, lookup)

    if writeMGH is True:
        if (
            INTEGRATE == "mean"
            or INTEGRATE == "median"
            or INTEGRATE == "mode"
            or INTEGRATE == "max"
            or INTEGRATE == "min"
        ):
            OUT_MGH_INTEGR = IN_SURF.replace(".vtk", "." + IN_SUFFIX + "-integrated.mgh")
            nb.freesurfer.save(
                nb.freesurfer.MGHImage(dataobj=np.asarray(integr)[:, 2].astype("float32"), affine=None),
                filename=OUT_MGH_INTEGR,
            )
        else:
            OUT_MGH = IN_SURF.replace(".vtk", "." + IN_SUFFIX + ".mgh")
            nb.freesurfer.save(
                nb.freesurfer.MGHImage(dataobj=lookup.astype("float32"), affine=None),
                filename=OUT_MGH,
            )

    if writeANNOT is True:
        OUT_ANNOT = IN_SURF.replace(".vtk", "." + IN_SUFFIX + ".annot")
        # ctab = np.array([(63,63,63,255,0), (255,0,0,255,234), (0,255,0,255,236), (0,0,255,255,238), (255,255,0,255,240), (255,255,0,255,246)])
        # names = ['Void', 'PrSbc', 'Sbc', 'CA1', 'CA2/3', 'ML']
        # labels = np.zeros(len(lookup))
        # labels[lookup==234] = 1
        # labels[lookup==236] = 2
        # labels[lookup==238] = 3
        # labels[lookup==240] = 4
        # labels[lookup==246] = 5
        # labels = labels.astype("int")
        # nb.freesurfer.write_annot(OUT_ANNOT, labels=labels, ctab=ctab, names=names, fill_ctab=False)

    # --------------------------------------------------------------------------
    # return

    return params


# ------------------------------------------------------------------------------
# CLI
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# MAIN PART

if __name__ == "__main__":
    # Command line options and error checking

    options = _parse_arguments()

    options = _check_arguments(options)

    # Run analysis

    mapValues(
        params=options.params,
        IN_VOL=options.IN_VOL,
        IN_SURF=options.IN_SURF,
        IN_LABEL=options.IN_LABEL,
        IN_INDICES=options.IN_INDICES,
        IN_SUFFIX=options.IN_SUFFIX,
        writePSOL=options.writePSOL,
        writeMGH=options.writeMGH,
        writeANNOT=options.writeANNOT,
        INTEGRATE=options.INTEGRATE,
        SELECT=options.SELECT,
        INTERP=options.INTERP,
    )
